import snapatac2 as snap

import numpy as np
import anndata as ad
from pathlib import Path
from collections import defaultdict
import pytest
from hypothesis import given, settings, HealthCheck, strategies as st
from hypothesis.extra.numpy import *
from scipy.sparse import csr_matrix
import gzip

from pytest import fixture
import os
import shutil

@fixture
def datadir(tmpdir, request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)

    if os.path.isdir(test_dir):
        for item in os.listdir(test_dir):
            src = os.path.join(test_dir, item)
            dst = os.path.join(tmpdir, item)
            if os.path.isdir(src):
                shutil.copytree(src, dst)
            else:
                shutil.copy2(src, dst)
    return tmpdir


def h5ad(dir=Path("./")):
    import uuid
    dir.mkdir(exist_ok=True)
    return str(dir / Path(str(uuid.uuid4()) + ".h5ad"))

def test_aggregation1():
    x = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float64)
    x = np.concat([x] * 3333, axis=0)

    groups = np.random.choice(["A", "B", "C", "D", "E"], size=x.shape[0])
    var_names_len = np.random.randint(1000, 5000, size=x.shape[1])
    var_names = [f"chr1:0-{i}" for i in var_names_len]

    adata = ad.AnnData(
        X=x,
        obs=dict(groups=groups),
    )
    adata.var_names = var_names

    expected = defaultdict(list)
    for g, v in zip(groups, list(x)):
        expected[g].append(v)
    for k in expected.keys():
        expected[k] = np.array(expected[k], dtype="float64").sum(axis=0)

    actual = snap.tl.aggregate_X(adata, groupby=groups)
    actual = {actual.obs_names[i]: np.ravel(actual.X[i, :]) for i in range(actual.n_obs)}
    for k, v in expected.items():
        np.testing.assert_array_almost_equal_nulp(v, actual[k])

    actual = snap.tl.aggregate_X(adata, groupby=groups, normalize='RPM')
    actual = {actual.obs_names[i]: np.ravel(actual.X[i, :]) for i in range(actual.n_obs)}
    for k, v in expected.items():
        v /= v.sum() / 1e6
        np.testing.assert_array_almost_equal_nulp(v, actual[k])

    actual = snap.tl.aggregate_X(adata, groupby=groups, normalize='RPKM')
    actual = {actual.obs_names[i]: np.ravel(actual.X[i, :]) for i in range(actual.n_obs)}
    for k, v in expected.items():
        v /= v.sum() / 1e6
        v /= (var_names_len / 1000.0)
        np.testing.assert_array_almost_equal_nulp(v, actual[k], nulp=3)

def test_aggregation2():
    x = np.random.poisson(1.0, (10_000, 50)).astype(np.float64)
    groups = np.random.choice(["A", "B", "C", "D", "E"], size=x.shape[0])
    obs_names = [str(i) for i in range(len(groups))]
    var_names = [str(i) for i in range(x.shape[1])]

    adata = ad.AnnData(
        X=x,
        obs=dict(ident=obs_names, groups=groups),
        var=dict(ident=var_names),
    )

    expected = defaultdict(list)
    for g, v in zip(groups, list(x)):
        expected[g].append(v)
    for k in expected.keys():
        expected[k] = np.array(expected[k], dtype="float64").sum(axis=0)
    expected = dict(expected.items())

    np.testing.assert_array_almost_equal_nulp(
        x.sum(axis=0),
        np.ravel(snap.tl.aggregate_X(adata).X),
    )

    np.testing.assert_array_almost_equal_nulp(
        np.array(list(expected.values())),
        snap.tl.aggregate_X(adata, groupby=groups).X,
    )

    adata.X = csr_matrix(adata.X[:])
    np.testing.assert_array_almost_equal_nulp(
        np.array(list(expected.values())),
        snap.tl.aggregate_X(adata, groupby=groups).X,
    )

def test_make_fragment(datadir, tmp_path):
    bam = str(datadir.join('test.bam'))
    bed = str(datadir.join('test.bed.gz'))
    output = str(tmp_path) + "/out.bed.gz"
    snap.pp.make_fragment_file(bam, output, True, barcode_regex="(^[ATCG]+):", chunk_size=5000)

    with gzip.open(bed, 'rt') as fl:
        expected = sorted(fl.readlines())

    with gzip.open(output, 'rt') as fl:
        actual = sorted(fl.readlines())
    
    assert expected == actual

@given(
    mat = arrays(
        np.float64, (50, 100),
        elements = {"allow_subnormal": False, "allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 100},
    ),
)
@settings(deadline = None, max_examples=5, suppress_health_check = [HealthCheck.function_scoped_fixture])
def test_reproducibility(mat):
    adata = ad.AnnData(X=csr_matrix(mat))
    embeddings = []
    for _ in range(3):
        embeddings.append(snap.tl.spectral(adata, features=None, random_state=0, inplace=False)[1])
    for x in embeddings:
        np.testing.assert_array_equal(x, embeddings[0])

    snap.tl.spectral(adata, features=None, random_state=0)
    knn = []
    for _ in range(3):
        knn.append(snap.pp.knn(adata, random_state=0, n_neighbors=25, inplace=False).todense())
    for x in knn:
        np.testing.assert_array_equal(x, knn[0])

    snap.pp.knn(adata, random_state=0, n_neighbors=25)
    leiden = []
    for _ in range(3):
        leiden.append(snap.tl.leiden(adata, random_state=0, resolution=1, n_iterations=10, inplace=False))
    for x in leiden:
        np.testing.assert_array_equal(x, leiden[0])

def read_bed(bed_file):
    content = []
    with gzip.open(bed_file, 'rt') as f:
        for line in f:
            if line.startswith('chr'):
                items = line.strip().split('\t')
                items.pop(4)
                content.append(items)
        return sorted(content)

def test_import(datadir):
    test_files = [
        (snap.datasets.pbmc500(downsample=True), True),
        (str(datadir.join('test_clean.tsv.gz')), True),
        (str(datadir.join('test_single.tsv.gz')), False),
    ]

    for fl, paired in test_files:
        data = snap.pp.import_fragments(
            fl,
            chrom_sizes=snap.genome.hg38,
            is_paired=paired,
            min_num_fragments=0,
            sorted_by_barcode=False,
        )

        data.obs['group'] = 'test_import'
        outputs = snap.ex.export_fragments(data, groupby="group", out_dir=str(datadir), suffix='.bed.gz')

        gold = read_bed(fl)
        test = read_bed(list(outputs.values())[0])
        for i, (g, t) in enumerate(zip(gold, test)):
            assert g == t, f"Line {i} mismatch: {g} != {t}"

def test_tile_matrix(datadir):
    def total_count(adata, bin_size):
        return snap.pp.add_tile_matrix(
            adata,
            bin_size=bin_size,
            inplace=False,
            counting_strategy='insertion',
            ).X.sum()

    data = snap.pp.import_fragments(
        snap.datasets.pbmc500(downsample=True),
        chrom_sizes=snap.genome.hg38,
        min_num_fragments=0,
        sorted_by_barcode=False,
    )

    counts = [total_count(data, i) for i in [500, 1000, 5000, 10000]]
    for i in range(1, len(counts)):
        assert counts[i] == counts[i - 1], f"Bin size {i} failed"

def test_spectral():
    fragment_file = snap.datasets.pbmc500(downsample=True)
    data = snap.pp.import_fragments(
        fragment_file,
        chrom_sizes=snap.genome.hg38,
        sorted_by_barcode=False,
    )
    snap.pp.add_tile_matrix(data)
    snap.metrics.tsse(data, snap.genome.hg38)
    snap.pp.filter_cells(data)
    snap.pp.select_features(data, n_features=4000)
    sp1 = snap.tl.spectral(data, random_state=0, inplace=False)[0]
    sp2 = snap.tl.spectral(data, random_state=0, inplace=False)[0]
    np.testing.assert_array_equal(sp1, sp2)
 


from __future__ import annotations

from pathlib import Path
from snapatac2._snapatac2 import AnnData, AnnDataSet
import snapatac2._snapatac2 as _snapatac2
import logging
from snapatac2.genome import Genome


def macs3(
    adata: AnnData | AnnDataSet,
    *,
    groupby: str | list[str] | None = None,
    qvalue: float = 0.05,
    call_broad_peaks: bool = False,
    broad_cutoff: float = 0.1,
    replicate: str | list[str] | None = None,
    replicate_qvalue: float | None = None,
    max_frag_size: int | None = None,
    selections: set[str] | None = None,
    nolambda: bool = False,
    shift: int = -100,
    extsize: int = 200,
    min_len: int | None = None,
    blacklist: Path | None = None,
    key_added: str = "macs3",
    tempdir: Path | None = None,
    inplace: bool = True,
    n_jobs: int = 8,
) -> dict[str, "polars.DataFrame"] | None:
    """Call peaks using MACS3.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    groupby
        Group the cells before peak calling. If a `str`, groups are obtained from
        `.obs[groupby]`. If None, peaks will be called for all cells.
    qvalue
        qvalue cutoff used in MACS3.
    call_broad_peaks
        If True, MACS3 will call broad peaks. The broad peak calling process
        utilizes two distinct cutoffs to discern broader, weaker peaks (`broad_cutoff`)
        and narrower, stronger peaks (`qvalue`), which are subsequently nested to
        provide a detailed peak landscape. To conceptualize "nested" peaks, picture
        a gene structure housing regions analogous to exons (strong peaks) and
        introns coupled with UTRs (weak peaks). Please note that, if you only want to
        call "broader" peak and not interested in the nested peak structure, please
        simply use `qvalue` with weaker cutoff instead of using `call_broad_peaks` option.
    broad_cutoff
        qvalue cutoff used in MACS3 for calling broad peaks.
    replicate
        Replicate information. If provided, reproducible peaks will be called
        for each group.
    replicate_qvalue
        qvalue cutoff used in MACS3 for calling peaks in replicates.
        This parameter is only used when `replicate` is provided.
        Typically this parameter is used to call peaks in replicates with a more lenient cutoff.
        If not provided, `qvalue` will be used.
    max_frag_size
        Maximum fragment size. If provided, fragments with sizes larger than
        `max_frag_size` will be not be used in peak calling.
        This is used in ATAC-seq data to remove fragments that are not
        from nucleosome-free regions.
        You can use :func:`~snapatac2.pl.frag_size_distr` to choose a proper value for
        this parameter.
    selections
        Call peaks for the selected groups only.
    nolambda
        If True, macs3 will use the background lambda as local lambda.
        This means macs3 will not consider the local bias at peak candidate regions.
    shift
        The shift size in MACS.
    extsize
        The extension size in MACS.
    min_len
        The minimum length of a called peak. If None, it is set to `extsize`.
    blacklist
        Path to the blacklist file in BED format. If provided, regions in the blacklist will be
        removed.
    key_added
        `.uns` key under which to add the peak information.
    tempdir
        If provided, a temporary directory will be created in the directory.
        Otherwise, a temporary directory will be created in the system default temporary directory.
    inplace
        Whether to store the result inplace.
    n_jobs
        Number of processes to use for peak calling.

    Returns
    -------
    dict[str, 'polars.DataFrame'] | None
        If `inplace=True` it stores the result in `adata.uns[`key_added`]`.
        Otherwise, it returns the result as dataframes.

    See Also
    --------
    merge_peaks
    """
    from MACS3.Signal.PeakDetect import PeakDetect
    from math import log
    import tempfile

    if isinstance(groupby, str):
        groupby = list(adata.obs[groupby])
    if replicate is not None and isinstance(replicate, str):
        replicate = list(adata.obs[replicate])

    # MACS3 options
    options = type("MACS3_OPT", (), {})()
    options.info = lambda _: None
    options.debug = lambda _: None
    options.warn = logging.warn
    options.name = "MACS3"
    options.bdg_treat = "t"
    options.bdg_control = "c"
    options.cutoff_analysis = False
    options.cutoff_analysis_file = "a"
    options.store_bdg = False
    options.do_SPMR = False
    options.trackline = False
    options.log_pvalue = None
    options.log_qvalue = log(qvalue, 10) * -1
    options.PE_MODE = False

    options.gsize = adata.uns["reference_sequences"][
        "reference_seq_length"
    ].sum()  # Estimated genome size
    options.maxgap = (
        30  # The maximum allowed gap between two nearby regions to be merged
    )
    options.minlen = extsize if min_len is None else min_len
    options.shift = shift
    options.nolambda = nolambda
    options.smalllocal = 1000
    options.largelocal = 10000
    options.call_summits = False if call_broad_peaks else True
    options.broad = call_broad_peaks
    if options.broad:
        options.log_broadcutoff = log(broad_cutoff, 10) * -1

    options.fecutoff = 1.0
    options.d = extsize
    options.scanwindow = 2 * options.d

    if groupby is None:
        peaks = _snapatac2.call_peaks_bulk(adata, options, max_frag_size)
        if inplace:
            adata.uns[key_added + "_pseudobulk"] = (
                peaks.to_pandas() if not adata.isbacked else peaks
            )
            return
        else:
            return peaks

    with tempfile.TemporaryDirectory(dir=tempdir) as tmpdirname:
        logging.info("Exporting fragments...")
        group_names = list(set(groupby))
        group_idx = {g: str(i) for i, g in enumerate(group_names)}
        fragments = _snapatac2.export_tags(
            adata,
            tmpdirname,
            [group_idx[x] for x in groupby],
            replicate,
            max_frag_size,
            selections,
        )

        def _call_peaks(tags):
            import tempfile

            tempfile.tempdir = tmpdirname  # Overwrite the default tempdir in MACS3
            merged, reps = _snapatac2.create_fwtrack_obj(tags)
            options.log_qvalue = log(qvalue, 10) * -1
            logging.getLogger().setLevel(
                logging.CRITICAL + 1
            )  # temporarily disable logging
            peakdetect = PeakDetect(treat=merged, opt=options)
            peakdetect.call_peaks()
            peakdetect.peaks.filter_fc(fc_low=options.fecutoff)
            merged = peakdetect.peaks

            others = []
            if replicate_qvalue is not None:
                options.log_qvalue = log(replicate_qvalue, 10) * -1
            for x in reps:
                peakdetect = PeakDetect(treat=x, opt=options)
                peakdetect.call_peaks()
                peakdetect.peaks.filter_fc(fc_low=options.fecutoff)
                others.append(peakdetect.peaks)

            logging.getLogger().setLevel(logging.INFO)  # enable logging
            return _snapatac2.find_reproducible_peaks(merged, others, blacklist)

        logging.info("Calling peaks...")
        if n_jobs == 1:
            peaks = [_call_peaks(x) for x in fragments.values()]
        else:
            peaks = _par_map(_call_peaks, [(x,) for x in fragments.values()], n_jobs)
        peaks = {group_names[int(k)]: v for k, v in zip(fragments.keys(), peaks)}
        if inplace:
            if adata.isbacked:
                adata.uns[key_added] = peaks
            else:
                adata.uns[key_added] = {k: v.to_pandas() for k, v in peaks.items()}
        else:
            return peaks


def merge_peaks(
    peaks: dict[str, "polars.DataFrame"],
    chrom_sizes: dict[str, int] | Genome,
    half_width: int = 250,
) -> "polars.DataFrame":
    """Merge peaks from different groups.

    Merge peaks from different groups. It is typically used to merge
    results from :func:`~snapatac2.tools.macs3`.

    This function initially expands the summits of identified peaks by `half_width`
    on both sides. Following this expansion, it addresses the issue of overlapping
    peaks through an iterative process. The procedure begins by prioritizing the
    most significant peak, determined by the smallest p-value. This peak is retained,
    and any peak that overlaps with it is excluded. Subsequently, the same method
    is applied to the next most significant peak. This iteration continues until
    all peaks have been evaluated, resulting in a final list of non-overlapping
    peaks, each with a fixed width determined by the initial extension.

    Parameters
    ----------
    peaks
        Peak information from different groups.
    chrom_sizes
        Chromosome sizes. If a :class:`~snapatac2.genome.Genome` is provided,
        chromosome sizes will be obtained from the genome.
    half_width
        Half width of the merged peaks.

    Returns
    -------
    'polars.DataFrame'
        A dataframe with merged peaks.

    See Also
    --------
    macs3
    """
    import pandas as pd
    import polars as pl

    chrom_sizes = (
        chrom_sizes.chrom_sizes if isinstance(chrom_sizes, Genome) else chrom_sizes
    )
    peaks = {
        k: pl.from_pandas(v) if isinstance(v, pd.DataFrame) else v
        for k, v in peaks.items()
    }
    return _snapatac2.py_merge_peaks(peaks, chrom_sizes, half_width)


def _par_map(mapper, args, nprocs):
    import time
    from multiprocess import get_context
    from tqdm import tqdm

    with get_context("spawn").Pool(nprocs) as pool:
        procs = set(pool._pool)
        jobs = [(i, pool.apply_async(mapper, x)) for i, x in enumerate(args)]
        results = []
        with tqdm(total=len(jobs)) as pbar:
            while len(jobs) > 0:
                if any(map(lambda p: not p.is_alive(), procs)):
                    raise RuntimeError("Some worker process has died unexpectedly.")

                remaining = []
                for i, job in jobs:
                    if job.ready():
                        results.append((i, job.get()))
                        pbar.update(1)
                    else:
                        remaining.append((i, job))
                jobs = remaining
                time.sleep(0.5)
        return [x for _, x in sorted(results, key=lambda x: x[0])]



from ._import_data import *
from ._basic import *
from ._knn import knn
from ._mnn_correct import mnc_correct
from ._harmony import harmony
from ._scanorama import scanorama_integrate
from ._scrublet import scrublet, filter_doublets
from ._recipe import *


# -- Path setup --------------------------------------------------------------

import re
import sys
import warnings
import os
import subprocess

import snapatac2

# -- Software version --------------------------------------------------------

# The short X.Y version (including .devXXXX, rcX, b1 suffixes if present)
version = re.sub(r'(\d+\.\d+)\.\d+(.*)', r'\1\2', snapatac2.__version__)
version = re.sub(r'(\.dev\d+).*?$', r'\1', version)

# The full version, including alpha/beta/rc tags.
release = snapatac2.__version__

# pyData/Sphinx-Theme version switcher
if ".dev" in version:
    switcher_version = "dev"
else:
    switcher_version = f"{version}"

print(f'Building documentation for SnapATAC2 {release} (short version: {version}, switcher version: {switcher_version})')

# -- Project information -----------------------------------------------------

project = 'SnapATAC2'
copyright = '2022-2025, Kai Zhang'
author = 'Kai Zhang'

# -- General configuration ---------------------------------------------------

suppress_warnings = ['ref.citation']
default_role = 'code'
add_function_parentheses = False

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
    "sphinx_autodoc_typehints",
    "sphinx_plotly_directive",
]

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

myst_enable_extensions = [
    "amsmath",
    #"colon_fence",
    #"deflist",
    "dollarmath",
    #"fieldlist",
    #"html_admonition",
    #"html_image",
    #"linkify",
    #"replacements",
    #"smartquotes",
    #"strikethrough",
    #"substitution",
    #"tasklist",
]

# Generate the API documentation when building
autosummary_generate = True
autodoc_member_order = 'bysource'
# autodoc_default_flags = ['members']
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True  # having a separate entry generally helps readability
napoleon_use_param = True
napoleon_custom_sections = [('Params', 'Parameters')]
todo_include_todos = False

intersphinx_mapping = {
    "anndata": ("https://anndata.readthedocs.io/en/stable/", None),
    "ipython": ("https://ipython.readthedocs.io/en/stable/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "scanpy": ("https://scanpy.readthedocs.io/en/stable/", None),
}

# Configure autodoc to handle import errors gracefully
autodoc_mock_imports = []

def resolve_py_module(name):
    """Resolve Python module aliases used in documentation."""
    import sys
    if name == 'pl':
        return sys.modules.get('snapatac2.plotting')
    elif name == 'tl':
        return sys.modules.get('snapatac2.tools')
    elif name == 'pp':
        return sys.modules.get('snapatac2.preprocessing')
    elif name == 'ex':
        return sys.modules.get('snapatac2.export')
    return None

def setup(app):
    """Configure Sphinx application."""
    # Register the aliases in sys.modules for Sphinx to find them
    import snapatac2
    import sys
    
    sys.modules['snapatac2.pl'] = snapatac2.pl
    sys.modules['snapatac2.tl'] = snapatac2.tl
    sys.modules['snapatac2.pp'] = snapatac2.pp
    sys.modules['snapatac2.ex'] = snapatac2.ex

smv_branch_whitelist = r'main'  # Include all branches

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

html_theme = 'pydata_sphinx_theme'
html_show_sphinx = False
html_show_sourcelink = False
html_static_path = ['_static']
html_css_files = [
    'css/custom.css',
]

html_theme_options = {
    "logo": {
        "text": "SnapATAC2",
        #"image_dark": "_static/logo-dark.svg",
        "alt_text": "SnapATAC2",
    },

    "github_url": "https://github.com/scverse/SnapATAC2",
    "external_links": [
        {"name": "Learn", "url": "https://kzhang.org/epigenomics-analysis/"}
    ],
    "header_links_before_dropdown": 6,

    "navbar_center": ["version-switcher", "navbar-nav"],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "navbar_align": "left",
    "show_version_warning_banner": switcher_version == "dev",

    "switcher": {
        "version_match": switcher_version,
        "json_url": "https://raw.githubusercontent.com/scverse/SnapATAC2/refs/heads/main/docs/_static/versions.json",
    },
}

commit = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('ascii')
code_url = f"https://github.com/scverse/SnapATAC2/blob/{commit}"

# based on numpy doc/source/conf.py
def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    import inspect

    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            with warnings.catch_warnings():
                # Accessing deprecated objects will generate noisy warnings
                warnings.simplefilter("ignore", FutureWarning)
                obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        try:  # property
            fn = inspect.getsourcefile(inspect.unwrap(obj.fget))
        except (AttributeError, TypeError):
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except TypeError:
        try:  # property
            source, lineno = inspect.getsourcelines(obj.fget)
        except (AttributeError, TypeError):
            lineno = None
    except OSError:
        lineno = None

    if lineno:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(snapatac2.__file__))

    return f"{code_url}/snapatac2-python/python/snapatac2/{fn}{linespec}"


from __future__ import annotations

import logging
from typing import Callable, Literal
from pathlib import Path
import numpy as np
import rustworkx as rx
import scipy.sparse as sp

from snapatac2.genome import Genome
from snapatac2._utils import fetch_seq
from snapatac2._snapatac2 import (
    AnnData, AnnDataSet, link_region_to_gene, PyDNAMotif, spearman
)

__all__ = ['NodeData', 'LinkData',
           'init_network_from_annotation', 'add_cor_scores', 'add_regr_scores',
           'add_tf_binding', 'link_tf_to_gene', 'prune_network', 'pagerank']

class NodeData:
    def __init__(self, id: str = "", type: str = "") -> None:
        self.id = id
        self.type = type
        self.regr_fitness = None

    def __repr__(self):
        return str(self.__dict__)

class LinkData:
    def __init__(
        self,
        distance: int =0,
        label: str | None = None,
    ) -> None:
        self.distance = distance
        self.label = label
        self.regr_score = None
        self.cor_score = None
    
    def __repr__(self):
        return str(self.__dict__)
   
def init_network_from_annotation(
    regions: list[str],
    anno_file: Path | Genome,
    upstream: int = 250000,
    downstream: int = 250000,
    id_type: Literal["gene_name", "gene_id", "transcript_id"] = "gene_name",
    coding_gene_only: bool = True,
) -> rx.PyDiGraph:
    """
    Build CRE-gene network from gene annotations.

    Link CREs to genes if they are close to genes' promoter regions.

    Parameters
    ----------
    regions
        A list of peaks/regions, e.g., `["chr1:100-1000", "chr2:55-222"]`.
    anno_file
        The GFF file containing the transcript level annotations.
    upstream
        Upstream extension to the transcription start site.
    downstream
        Downstream extension to the transcription start site.
    id_type
        "gene_name", "gene_id" or "transcript_id".
    coding_gene_only
        Retain only coding genes in the network.

    Returns
    -------
    rx.PyDiGraph:
        A network where peaks/regions point towards genes if they are within genes'
        regulatory domains.
    """
    if isinstance(anno_file, Genome):
        anno_file = anno_file.annotation
        
    region_added = {}
    graph = rx.PyDiGraph()
    links = link_region_to_gene(
        regions,
        str(anno_file),
        upstream,
        downstream,
        id_type,
        coding_gene_only,
    )
    for (id, type), regions in links.items():
        to = graph.add_node(NodeData(id.upper(), type))
        for i, t, distance in regions:
            key = (i, t)
            if key in region_added:
                graph.add_edge(region_added[key], to, LinkData(distance))
            else:
                region_added[key] = graph.add_parent(to, NodeData(i, t), LinkData(distance))
    return graph

def add_cor_scores(
    network: rx.PyDiGraph,
    *,
    gene_mat: AnnData | AnnDataSet | None = None,
    peak_mat: AnnData | AnnDataSet | None = None,
    select: list[str] | None = None,
    overwrite: bool = False,
):
    """
    Compute correlation scores for any two connected nodes in the network.

    This function can be used to compute correlation scores for any type of
    associations. There are typically three types of edges in the network:

    1. Region -> gene: CREs regulate target genes.
    2. Gene -> gene: genes regulate other genes.
    3. Gene -> region: TFs bind to CREs.

    Parameters
    ----------
    network
        network
    gene_mat
        AnnData or AnnDataSet object storing the cell by gene count matrix,
        where the `.var_names` contains genes.
    peak_mat
        AnnData or AnnDataSet object storing the cell by peak count matrix,
        where the `.var_names` contains peaks.
    select
        Run this for selected genes only.
    overwrite
        Whether to overwrite existing records.
    """
    from tqdm import tqdm

    key = "cor_score"
    if list(peak_mat.obs_names) != list(gene_mat.obs_names):
        raise NameError("gene matrix and peak matrix should have the same obs_names")
    if select is not None:
        select = set(select)
    without_overwrite = None if overwrite else key 

    if network.num_edges() > 0:
        data = _get_data_iter(network, peak_mat, gene_mat, select, without_overwrite)
        for (nd_X, X), (nd_y, y) in tqdm(data):
            if sp.issparse(X):
                X = X.todense()
            if sp.issparse(y):
                y = y.todense()
            scores = np.ravel(spearman(X.T, y.reshape((1, -1))))
            for nd, sc in zip(nd_X, scores):
                setattr(network.get_edge_data(nd, nd_y), key, sc)

def add_regr_scores(
    network: rx.PyDiGraph,
    *,
    peak_mat: AnnData | AnnDataSet | None = None,
    gene_mat: AnnData | AnnDataSet | None = None,
    select: list[str] | None = None,
    method: Literal["gb_tree", "elastic_net"] = "elastic_net",
    scale_X: bool = False,
    scale_Y: bool = False,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    use_gpu: bool = False,
    overwrite: bool = False,
):
    """
    Perform regression analysis for nodes and their parents in the network.

    Parameters
    ----------
    network
        network
    peak_mat
        AnnData or AnnDataSet object storing the cell by peak count matrix,
        where the `.var_names` contains peaks.
    gene_mat
        AnnData or AnnDataSet object storing the cell by gene count matrix,
        where the `.var_names` contains genes.
    select
        Run this for selected genes only.
    method
        Regresson model.
    scale_X
        Whether to scale the features.
    scale_Y
        Whether to scale the response variable.
    alpha
        Constant that multiplies the penalty terms in 'elastic_net'.
    l1_ratio
        Used in 'elastic_net'. The ElasticNet mixing parameter,
        with `0 <= l1_ratio <= 1`. For `l1_ratio = 0` the penalty is an L2 penalty.
        For `l1_ratio = 1` it is an L1 penalty. For `0 < l1_ratio < 1`,
        the penalty is a combination of L1 and L2.
    use_gpu
        Whether to use gpu
    overwrite
        Whether to overwrite existing records.
    """
    from tqdm import tqdm

    key = "regr_score"
    if peak_mat is not None and gene_mat is not None:
        if list(peak_mat.obs_names) != list(gene_mat.obs_names):
            raise NameError("gene matrix and peak matrix should have the same obs_names")
    if select is not None:
        select = set(select)
    without_overwrite = None if overwrite else key 
    tree_method = "gpu_hist" if use_gpu else "hist"
    
    if network.num_edges() == 0:
        return network

    for (nd_X, X), (nd_y, y) in tqdm(_get_data_iter(network, peak_mat, gene_mat, select, without_overwrite, scale_X, scale_Y)):
        y = np.ravel(y.todense()) if sp.issparse(y) else y
        if method == "gb_tree":
            scores, fitness = _gbTree(X, y, tree_method=tree_method)
        elif method == "elastic_net":
            scores, fitness = _elastic_net(X, y, alpha, l1_ratio)
        elif method == "logistic_regression":
            scores, fitness = _logistic_regression(X, y)
        else:
            raise NameError("Unknown method")
        network[nd_y].regr_fitness = fitness
        for nd, sc in zip(nd_X, scores):
            setattr(network.get_edge_data(nd, nd_y), key, sc)

def add_tf_binding(
    network: rx.PyDiGraph,
    *,
    motifs: list[PyDNAMotif],
    genome_fasta: Path | Genome,
    pvalue: float = 1e-5,
):
    """Add TF motif binding information.

    Parameters
    ----------
    network
        Network
    motifs
        TF motifs
    genome_fasta
        A fasta file containing the genome sequences or a Genome object.
    pvalue
        P-value threshold for motif binding.
    """
    from pyfaidx import Fasta
    from tqdm import tqdm
    import itertools

    regions = [(i, network[i].id) for i in network.node_indices() if network[i].type == "region"]
    logging.info("Fetching {} sequences ...".format(len(regions)))
    genome = genome_fasta.fasta if isinstance(genome_fasta, Genome) else str(genome_fasta)
    genome = Fasta(genome, one_based_attributes=False)
    sequences = [fetch_seq(genome, region) for _, region in regions]

    logging.info("Searching for the binding sites of {} motifs ...".format(len(motifs)))
    for motif in tqdm(motifs):
        bound = motif.with_nucl_prob().exists(sequences, pvalue=pvalue)
        if any(bound):
            name = motif.id if motif.name is None else motif.name
            nid = network.add_node(NodeData(name.upper(), "motif"))
            network.add_edges_from(
                [(nid, i, LinkData()) for i, _ in itertools.compress(regions, bound)]
            )

def link_tf_to_gene(network: rx.PyDiGraph) -> rx.PyDiGraph:
    """Contruct a genetic network by linking TFs to target genes.
    
    Convert the network to a genetic network.
    :func:`~snapatac2.tl.add_tf_binding` must be ran first in order to use this function.

    Parameters
    ----------

    Returns 
    -------
    rx.PyDiGraph
    """
    def aggregate(edge_data):
        best = 0
        for a, b in edge_data:
            sc = min(abs(a.cor_score), abs(b.cor_score))
            if sc >= best:
                e1 = a
                e2 = b
        return (e1, e2)

    graph = rx.PyDiGraph()

    genes = {}
    for node in network.nodes():
        if node.type == "gene":
            genes[node.id] = graph.add_node(node)

    edges = []
    for nid in network.node_indices():
        node = network[nid]
        if node.type == "motif" and node.id in genes:
            fr = genes[node.id]
            targets = {}
            for succ_id in network.successor_indices(nid):
                target = network[succ_id]
                if target.type == "region":
                    edge_data = network.get_edge_data(nid, succ_id)
                    for gene_id in network.successor_indices(succ_id):
                        region_gene = network.get_edge_data(succ_id, gene_id)
                        to = genes[network[gene_id].id]
                        if to in targets :
                            targets[to].append((edge_data, region_gene))
                        else:
                            targets[to] = [(edge_data, region_gene)]

            for k, v in targets.items():
                    #e1, e2 = aggregate(v)
                    #label = ''.join(['+' if x.cor_score > 0 else '-' for x in [e1, e2]])
                    #edges.append((fr, k, LinkData(label=label)))
                    edges.append((fr, k, LinkData()))

    graph.add_edges_from(edges)

    remove = []
    for nid in graph.node_indices():
        if graph.in_degree(nid) + graph.out_degree(nid) == 0:
            remove.append(nid)
    if len(remove) > 0:
        graph.remove_nodes_from(remove)
    
    return graph

def prune_network(
    network: rx.PyDiGraph,
    node_filter: Callable[[NodeData], bool] | None = None,
    edge_filter: Callable[[int, int, LinkData], bool] | None = None,
    remove_isolates: bool = True,
) -> rx.PyDiGraph:
    """
    Prune the network.

    Parameters
    ----------
    network
        network
    node_filter
        Node filter function.
    edge_filter
        Edge filter function.
    remove_isolates
        Whether to remove isolated nodes.

    Returns
    -------
    rx.PyDiGraph
    """
    graph = rx.PyDiGraph()
    
    node_retained = [nid for nid in network.node_indices()
                     if node_filter is None or node_filter(network[nid])]              
    node_indices = graph.add_nodes_from([network[nid] for nid in node_retained])
    node_index_map = dict(zip(node_retained, node_indices))
   
    edge_retained = [(node_index_map[fr], node_index_map[to], data)
                     for fr, to, data in network.edge_index_map().values()
                     if fr in node_index_map and to in node_index_map and
                        (edge_filter is None or edge_filter(fr, to, data))]

    graph.add_edges_from(edge_retained)

    if remove_isolates:
        remove = []
        for nid in graph.node_indices():
            if graph.in_degree(nid) + graph.out_degree(nid) == 0:
                remove.append(nid)
        if len(remove) > 0:
            graph.remove_nodes_from(remove)
            logging.info("Removed {} isolated nodes.".format(len(remove)))

    return graph

def pagerank(
    network,
    node_weights: str | list[float] | None = None,
    edge_weights: str | list[float] | None = None,
) -> list[tuple[str, float]]:
    tfs = {network[nid].id for nid in network.node_indices() if network.out_degree(nid) > 0}
    g = _to_igraph(network, node_weights, edge_weights, True)
    pagerank_scores = g.personalized_pagerank(
        reset=None if node_weights is None else 'weight',
        weights=None if edge_weights is None else 'weight',
    )
    return [(i['name'], s) for i, s in zip(g.vs, pagerank_scores) if i['name'] in tfs]

def _to_igraph(
    graph,
    node_weights: str | list[float] | None = None,
    edge_weights: str | list[float] | None = None,
    reverse_edge: bool = False,
):
    import igraph as ig
    g = ig.Graph()

    nodes = [x.id for x in graph.nodes()]
    node_attributes = None
    if node_weights is not None:
        if isinstance(node_weights, str):
            node_attributes = {"weight": [getattr(x, node_weights) for x in graph.nodes()]}
        else:
            node_attributes = {"weight": node_weights}
    g.add_vertices(nodes, attributes=node_attributes)
    
    edges = []
    if edge_weights is not None and isinstance(edge_weights, list):
        weights = edge_weights
    else:
        weights = []
    for fr, to, edge in graph.edge_index_map().values():
        if reverse_edge:
            edges.append((graph[to].id, graph[fr].id))
        else:
            edges.append((graph[fr].id, graph[to].id))
        if edge_weights is not None and isinstance(edge_weights, str):
            weights.append(getattr(edge, edge_weights))
    if len(weights) > 0:
        edge_attributes = {"weight": weights}
    else:
        edge_attributes = None
    g.add_edges(edges, attributes=edge_attributes)

    return g

def _network_stats(network: rx.PyDiGraph):
    from collections import defaultdict

    nNodes = network.num_nodes()
    nEdges = network.num_edges()
    region_stat = defaultdict(lambda: {'nParents': defaultdict(lambda: 0), 'nChildren': defaultdict(lambda: 0)})
    motif_stat = defaultdict(lambda: {'nParents': defaultdict(lambda: 0), 'nChildren': defaultdict(lambda: 0)})
    gene_stat = defaultdict(lambda: {'nParents': defaultdict(lambda: 0), 'nChildren': defaultdict(lambda: 0)})

    for fr, to, data in network.edge_index_map().values():
        fr_type = network[fr].type
        to_type = network[to].type
 

class _DataPairIter:
    """
    Interator generating X and y pairs.

    ...

    Attributes
    ----------
    regulator_mat
        Regulator data.
    regulatee_mat
        Regulatee data.
    regulator_idx_map
        Node id to regulator matrix index map.
    regulator_ids
        Node ids of regulators.
    regulatee_ids
        Node ids of regulatees.
    """
    def __init__(
        self,
        mat_X,
        mat_Y,
        idx_map_X,
        id_XY,
    ) -> None:
        self.mat_X = mat_X
        self.mat_Y = mat_Y
        self.idx_map_X = idx_map_X
        self.id_XY = id_XY
        self.index = 0

    def __len__(self):
        return self.mat_Y.shape[1]

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        if self.index >= self.__len__():
            raise StopIteration

        nd_X, nd_y = self.id_XY[self.index]
        y = self.mat_Y[:, self.index]
        X = self.mat_X[:, [self.idx_map_X[nd] for nd in nd_X]]

        self.index += 1
        return (nd_X, X), (nd_y, y)

def _get_data_iter(
    network: rx.PyDiGraph,
    peak_mat: AnnData | AnnDataSet | None,
    gene_mat: AnnData | AnnDataSet | None,
    select: set[str] | None = None,
    without_overwrite: str | None = None,
    scale_X: bool = False,
    scale_Y: bool = False,
) -> _DataPairIter:
    """
    """
    from scipy.stats import zscore

    def get_mat(nids, node_getter, gene_mat, peak_mat):
        genes = []
        peaks = []
        mats = []

        for x in nids:
            nd = node_getter(x) 
            if nd.type == "gene" or nd.type == "motif":
                genes.append(x)
            elif nd.type == "region":
                peaks.append(x)
            else:
                raise NameError("unknown type: {}".format(nd.type))

        if len(genes) != 0:
            if len(genes) == gene_mat.n_vars:
                mats.append(gene_mat.X[:])
            else:
                idx_map = {x.upper(): i for i, x in enumerate(gene_mat.var_names)}
                ix = [idx_map[node_getter(x).id] for x in genes]
                mats.append(gene_mat.X[:, ix])

        if len(peaks) != 0:
            if len(peaks) == peak_mat.n_vars:
                mats.append(peak_mat.X[:])
            else:
                if peak_mat.isbacked:
                    ix = peak_mat.var_ix([node_getter(x).id for x in peaks])
                else:
                    ix = [peak_mat.var_names.get_loc(node_getter(x).id) for x in peaks]
                mats.append(peak_mat.X[:, ix])
        
        if all([sp.issparse(x) for x in mats]):
            mat = sp.hstack(mats, format="csc")
        else:
            mat = np.hstack(mats)
        return (genes + peaks, mat)

    all_genes = set([x.upper() for x in gene_mat.var_names])
    select = all_genes if select is None else select
    id_XY = []

    for nid in network.node_indices():
        if network[nid].type == "region" or network[nid].id in select:
            parents = [pid for pid, _, edge_data in network.in_edges(nid)
                       if (without_overwrite is None or
                           getattr(edge_data, without_overwrite) is None) and
                          (network[pid].type == "region" or
                           network[pid].id in all_genes)]
            if len(parents) > 0:
                id_XY.append((parents, nid))
    unique_X = list({y for x, _ in id_XY for y in x})

    id_XY, mat_Y = get_mat(id_XY, lambda x: network[x[1]], gene_mat, peak_mat)

    unique_X, mat_X = get_mat(unique_X, lambda x: network[x], gene_mat, peak_mat)

    if scale_X:
        if sp.issparse(mat_X):
            logging.warning("Try to scale a sparse matrix")
        mat_X = zscore(mat_X, axis=0)
    if scale_Y:
        if sp.issparse(mat_Y):
            logging.warning("Try to scale a sparse matrix")
        mat_Y = zscore(mat_Y, axis=0)
        
    return _DataPairIter(
        mat_X,
        mat_Y,
        {v: i for i, v in enumerate(unique_X)},
        id_XY,
    )

def _logistic_regression(X, y):
    from sklearn.linear_model import LogisticRegression 

    y = y != 0
    regr = LogisticRegression(max_iter=1000, random_state=0).fit(X, y)
    return np.ravel(regr.coef_), regr.score(X, y)

def _elastic_net(X, y, alpha=1, l1_ratio=0.5, positive=False):
    from sklearn.linear_model import ElasticNet

    X = np.asarray(X)
    y = np.asarray(y)

    regr = ElasticNet(
        alpha=alpha, l1_ratio=l1_ratio, positive=positive,
        random_state=0, copy_X=False, max_iter=10000,
    ).fit(X, y)
    return regr.coef_, regr.score(X, y)

def _gbTree(X, y, tree_method = "hist"):
    import xgboost as xgb
    regr = xgb.XGBRegressor(tree_method = tree_method).fit(X, y)
    return regr.feature_importances_, regr.score(X, y)


import numpy as np
import logging

from anndata import AnnData
import snapatac2._snapatac2 as internal

def is_anndata(data) -> bool:
    return isinstance(data, AnnData) or isinstance(data, internal.AnnData) or isinstance(data, internal.AnnDataSet)

def anndata_par(adatas, func, n_jobs=4):
    return anndata_ipar(list(enumerate(adatas)), lambda x: func(x[1]), n_jobs=n_jobs)

def anndata_ipar(inputs, func, n_jobs=4):
    from tqdm import tqdm
    
    exist_in_memory_adata = False
    for _, adata in inputs:
        if isinstance(adata, AnnData):
            exist_in_memory_adata = True
            break
    if exist_in_memory_adata:
        logging.warn(("Input contains in-memory AnnData objects. "
                      "Multiprocessing will not be used. "
                      "To enable multiprocessing, use backed AnnData objects"))
        return [func((i, adata)) for i, adata in tqdm(inputs)]
    else:
        from multiprocess import get_context

        def _func(x):
            adata = internal.read(x[1], backend=x[2])
            result = func((x[0], adata))
            adata.close() 
            return result

        # Close the AnnData objects and return the filenames
        files = []
        for i, adata in inputs:
            files.append((i, adata.filename, adata.backend))
            adata.close()

        with get_context("spawn").Pool(n_jobs) as p:
            result = list(tqdm(p.imap(_func, files), total=len(files)))
        
        # Reopen the files if they were closed
        for _, adata in inputs:
            adata.open(mode='r+')
        
        return result

def get_file_format(suffix):
    suffix = suffix.lower()
    _suffix = suffix

    if suffix.endswith(".gz"):
        compression = "gzip"
        _suffix = suffix[:-3]
    elif suffix.endswith(".zst"):
        compression = "zstandard"
        _suffix = suffix[:-4]
    else:
        compression = None
    
    if suffix.endswith(".bw") or suffix.endswith(".bigwig"):
        format = "bigwig"
    elif _suffix.endswith(".bedgraph") or _suffix.endswith(".bg") or _suffix.endswith(".bdg"):
        format = "bedgraph"
    else:
        format = None
    
    return format, compression

def get_igraph_from_adjacency(adj):
    """Get igraph graph from adjacency matrix."""
    import igraph as ig
    vcount = max(adj.shape)
    sources, targets = adj.nonzero()
    edgelist = list(zip(list(sources), list(targets)))
    weights = np.ravel(adj[(sources, targets)])
    gr = ig.Graph(n=vcount, edges=edgelist, directed=False, edge_attrs={"weight": weights})
    return gr

def chunks(mat, chunk_size: int):
    """
    Return chunks of the input matrix
    """
    n = mat.shape[0]
    for i in range(0, n, chunk_size):
        j = max(i + chunk_size, n)
        yield mat[i:j, :]

def find_elbow(x, saturation=0.01):
    accum_gap = 0
    for i in range(1, len(x)):
        gap = x[i-1] - x[i]
        accum_gap = accum_gap + gap
        if gap < saturation * accum_gap:
            return i
    return None

def fetch_seq(fasta, region):
    chr, x = region.split(':')
    start, end = x.split('-')
    start = int(start)
    end = int(end)
    seq = fasta[chr][start:end].seq
    l1 = len(seq)
    l2 = end - start
    if l1 != l2:
        raise NameError(
            "sequence fetch error: expected length: {}, but got {}.".format(l2, l1)
        )
    else:
        return seq

def pcorr(A, B):
    """Compute pairwsie correlation between two matrices.

    A
        n_sample x n_feature
    B
        n_sample x n_feature
    """
    N = B.shape[0]

    sA = A.sum(0)
    sB = B.sum(0)

    p1 = N * np.einsum('ij,ik->kj', A, B)
    p2 = sA * sB[:,None]
    p3 = N * ((B**2).sum(0)) - (sB**2)
    p4 = N * ((A**2).sum(0)) - (sA**2)

    return (p1 - p2) / np.sqrt(p4 * p3[:,None])


"""
Use harmony to integrate cells from different experiments.
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import multiprocessing as mp
from functools import partial

import snapatac2._snapatac2 as internal
from snapatac2._utils import is_anndata


def harmony(
    adata: internal.AnnData | internal.AnnDataSet | np.ndarray,
    *,
    batch: str | list[str],
    use_rep: str = "X_spectral",
    use_dims: int | list[int] | None = None,
    groupby: str | list[str] | None = None,
    key_added: str | None = None,
    inplace: bool = True,
    n_jobs: int = 8,
    **kwargs,
) -> np.ndarray | None:
    """
    Use harmonypy to integrate different experiments.

    Harmony is an algorithm for integrating single-cell
    data from multiple experiments. This function uses the python
    port of Harmony, ``harmonypy``, to integrate single-cell data
    stored in an AnnData object. This function should be run after performing
    dimension reduction.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    batch
        The name of the column in ``adata.obs`` that differentiates
        among experiments/batches.
    use_rep
        The name of the field in ``adata.obsm`` where the lower dimensional
        representation is stored.
    use_dims
        Use these dimensions in `use_rep`.
    groupby
        If specified, split the data into groups and perform batch correction
        on each group separately.
    key_added
        If specified, add the result to ``adata.obsm`` with this key. Otherwise,
        it will be stored in ``adata.obsm[use_rep + "_harmony"]``.
    inplace
        Whether to store the result in the anndata object.
    kwargs
        Any additional arguments will be passed to
        ``harmonypy.run_harmony()``.

    Returns
    -------
    np.ndarray | None
        if `inplace=True` it updates adata with the field
        ``adata.obsm[`use_rep`_harmony]``, containing principal components
        adjusted by Harmony such that different experiments are integrated.
        Otherwise, it returns the result as a numpy array.
    """
    # Check if the data is in an AnnData object
    if is_anndata(adata):
        mat = adata.obsm[use_rep]
    else:
        mat = adata
        inplace = False

    # Use only the specified dimensions
    if isinstance(use_dims, int):
        use_dims = range(use_dims)
    mat = mat if use_dims is None else mat[:, use_dims]

    # Create a pandas dataframe with the batch information
    if isinstance(batch, str) or isinstance(batch, list):
        batch = pd.DataFrame(adata.obs[batch])

    if groupby is None:
        mat = _harmony(mat, batch, **kwargs)
    else:
        if isinstance(groupby, str):
            groupby = adata.obs[groupby]
        groups = list(set(groupby))
        group_idxs = [
            [i for i, x in enumerate(groupby) if x == group] for group in groups
        ]
        margs = [
            (mat[group_idx, :], batch.iloc[group_idx, :].copy())
            for group_idx in group_idxs
        ]

        with mp.Pool(processes=min(n_jobs, len(groups))) as pool:
            mats = pool.starmap(partial(_harmony, **kwargs), margs)
        for i, group_idx in enumerate(group_idxs):
            mat[group_idx, :] = mats[i]

    if inplace:
        if key_added is None:
            adata.obsm[use_rep + "_harmony"] = mat
        else:
            adata.obsm[key_added] = mat
    else:
        return mat


def _harmony(data_matrix, batch_labels, **kwargs):
    try:
        import harmonypy
    except ImportError:
        raise ImportError("\nplease install harmonypy:\n\n\tpip install harmonypy")
    if data_matrix.shape[0] == 1:
        return data_matrix
    if len(batch_labels.shape) == 1:
        batch_labels = pd.DataFrame(batch_labels)
    # check if batch has >1 unique values
    for b in batch_labels.columns:
        if len(batch_labels[b].unique()) == 1:
            batch_labels = batch_labels.drop(b, axis=1)
    if batch_labels.shape[1] == 0:
        return data_matrix
    harmony_out = harmonypy.run_harmony(
        data_matrix,
        pd.DataFrame(batch_labels.values, columns=batch_labels.columns),
        batch_labels.columns.values,
        **kwargs,
    )
    return harmony_out.Z_corr.T



from setuptools import setup

from pathlib import Path

ROOT_DIR = Path(__file__).parent
README = (ROOT_DIR / "README.md").read_text()

VERSION = {}
with open(ROOT_DIR / "snapatac2_contrib/_version.py") as fp:
    exec(fp.read(), VERSION)

setup(
    name="snapatac2-contrib",
    description='SnapATAC2: Single-cell epigenomics analysis pipeline',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://scverse.org/SnapATAC2/', 
    license='MIT',
    version=VERSION['__version__'],
    packages=[
        "snapatac2_contrib",
        "snapatac2_contrib.metrics",
    ],
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[
        "snapatac2>=2.3.0",
        "anndata>=0.8.0",
        "numpy>=1.16.0",
        "pandas",
        "polars>=0.14.0, <=0.18.0",
        "scipy>=1.4",
        "typing_extensions",
    ],
)



from warnings import warn

import igraph as ig
import leidenalg as la
import networkx as nx
import numpy as np
import pandas as pd


def sdi(data):
    """
    Calculate the Shannon Diversity Index (SDI) for a categorical data series.

    Higher SDI indicates higher diversity.
    """
    if len(data) == 0:
        return 0
    counts = pd.Series(data).value_counts()
    ratio = counts / counts.sum()
    score = -(np.log(ratio) * ratio).sum()
    return score


def calculate_direct_confusion(*args, **kwargs):
    warn("This function is deprecated. Call calculate_overlap_score instead", DeprecationWarning)
    return calculate_overlap_score(*args, **kwargs)


def _get_overlap_score(left_values, right_values):
    score = np.min([left_values, right_values], axis=0).sum()
    return score


def calculate_overlap_score(left_part, right_part):
    """
    Calculate the overlap score between intra-dataset clusters using co-cluster information.

    Input are 2 dataframes for left/source and right/target dataset,
    Each dataframe only contain 2 columns, first is original cluster, second is co-cluster.
    The returned confusion matrix will be the form of source-cluster by target-cluster.

    Parameters
    ----------
    left_part
        Dataframe for left/source dataset.
    right_part
        Dataframe for right/target dataset.
    """
    left_part = left_part.astype(str)
    original_left_name = left_part.columns[0]
    left_part.columns = ["cluster", "co_cluster"]

    right_part = right_part.astype(str)
    original_right_name = right_part.columns[0]
    if original_right_name == original_left_name:
        original_right_name += "_1"
    right_part.columns = ["cluster", "co_cluster"]

    left_confusion = left_part.groupby("cluster")["co_cluster"].value_counts().unstack()
    right_confusion = right_part.groupby("cluster")["co_cluster"].value_counts().unstack()

    left_confusion_portion = left_confusion.divide(left_confusion.sum(axis=1), axis=0).fillna(0)
    right_confusion_portion = right_confusion.divide(right_confusion.sum(axis=1), axis=0).fillna(0)

    union_index = left_confusion_portion.columns.intersection(right_confusion_portion.columns)
    left_confusion_portion = left_confusion_portion.reindex(columns=union_index).fillna(0)
    right_confusion_portion = right_confusion_portion.reindex(columns=union_index).fillna(0)

    records = []
    for left_cluster, left_row in left_confusion_portion.iterrows():
        for right_cluster, right_row in right_confusion_portion.iterrows():
            overlap_value = _get_overlap_score(left_row.values, right_row.values)
            records.append([left_cluster, right_cluster, overlap_value])

    flat_confusion_matrix = pd.DataFrame(records, columns=[original_left_name, original_right_name, "overlap_value"])

    confusion_matrix = flat_confusion_matrix.pivot(
        index=original_left_name, columns=original_right_name, values="overlap_value"
    )
    return confusion_matrix


def calculate_diagonal_score(confusion_matrix, col_group, row_group):
    """
    Given a confusion matrix, evaluate the overall integration performance with the diagonal score.

    Parameters
    ----------
    confusion_matrix :
        A confusion matrix.
    col_group :
        Integration group for the columns.
    row_group :
        Integration group for the rows.

    Returns
    -------
    float
    """
    # group the confusion matrix by the col_group and row_group,
    # then for each block, calculate the mean overlap score
    group_overlap_score_mean = confusion_matrix.groupby(col_group, axis=1).apply(
        lambda sub_df: sub_df.groupby(row_group, axis=0).mean().mean(axis=1)
    )

    # diagonal score is the ratio of the diagonal overlap score sum to the non-diagonal overlap score sum
    diag_sum = np.diag(group_overlap_score_mean, 0).sum()
    non_diag_sum = group_overlap_score_mean.values.sum() - diag_sum

    ratio = diag_sum / non_diag_sum
    return ratio


def confusion_matrix_clustering(
    confusion_matrix, min_value=0.1, max_value=0.9, partition_type=None, resolution=1, seed=0
):
    """
    Given a confusion matrix, bi-clustering the matrix using Leiden Algorithm.

    Parameters
    ----------
    confusion_matrix :
        A confusion matrix. Row is query, column is reference.
    min_value :
        minimum value to be used as an edge weight.
    max_value :
        maximum value to be used as an edge weight. Larger value will be capped to this value.
    partition_type :
        The type of partition to be used. See leidenalg documentation for details.
        If None, use the :class:`~leidenalg.RBConfigurationVertexPartition` partition type.
    resolution :
        The resolution parameter to be used. See leidenalg documentation for details.
    seed :
        random seed for Leiden Algorithm.

    Returns
    -------
    query_group:
        A series of query cluster integration group.
    reference_group:
        A series of reference cluster integration group.
    confusion_matrix：
        A confusion matrix ordered by the integration group.
    g:
        The Graph object used for leiden clustering and determine integration groups.
    modularity_score:
        The modularity score of the graph partition by integration groups.
    """
    # make sure confusion matrix index and columns are unique
    try:
        assert confusion_matrix.index.duplicated().sum() == 0
        assert confusion_matrix.columns.duplicated().sum() == 0
    except AssertionError:
        raise ValueError("Confusion matrix index and columns should be unique")

    confusion_matrix = confusion_matrix.copy()
    # map id to int
    # row is query, column is reference
    query_idx_map = {c: i for i, c in enumerate(confusion_matrix.index)}
    ref_idx_map = {c: i + len(query_idx_map) for i, c in enumerate(confusion_matrix.columns)}
    confusion_matrix.index = confusion_matrix.index.map(query_idx_map)
    confusion_matrix.columns = confusion_matrix.columns.map(ref_idx_map)

    # build a weighted graph from sig scores
    edges = confusion_matrix.unstack()
    edges = edges[edges > min_value].copy().reset_index()
    edges.columns = ["ref", "query", "weight"]
    g = nx.Graph()
    for _, (ref, query, weight) in edges.iterrows():
        weight = min(max_value, weight)
        g.add_edge(ref, query, weight=weight)
        # convert to igraph to run leiden
    h = ig.Graph.from_networkx(g)

    if partition_type is None:
        partition_type = la.RBConfigurationVertexPartition

    # leiden clustering
    partition = la.find_partition(h, partition_type, weights="weight", seed=seed, resolution_parameter=resolution)

    # store output into adata.obs
    groups = np.array(partition.membership)
    idx_to_group = {int(node): g for node, g in zip(g.nodes, groups)}

    # map cluster back to original labels
    # -1 means a ref or query cluster is not co-clustered with the other dataset,
    # so the OS is very small and not included in the graph
    query_group = pd.Series(query_idx_map).map(idx_to_group).fillna(-1).astype(int)
    ref_group = pd.Series(ref_idx_map).map(idx_to_group).fillna(-1).astype(int)

    # also make an ordered confusion matrix
    confusion_matrix.index = confusion_matrix.index.map({v: k for k, v in query_idx_map.items()})
    confusion_matrix.columns = confusion_matrix.columns.map({v: k for k, v in ref_idx_map.items()})
    confusion_matrix = confusion_matrix.loc[query_group.sort_values().index, ref_group.sort_values().index].copy()

    # calculate the graph modularity after partition
    from collections import defaultdict

    from networkx.algorithms.community import modularity

    group_nodes = defaultdict(set)
    for group, node in zip(groups, g.nodes):
        group_nodes[group].add(node)
    modularity_score = modularity(g, group_nodes.values())
    return query_group, ref_group, confusion_matrix, g, modularity_score



from sys import stderr
from ._io import read_10x_mtx
from importlib.metadata import version
from . import preprocessing as pp
from . import tools as tl
from . import metrics
from . import plotting as pl
from . import export as ex

from snapatac2._snapatac2 import (
    set_write_options, get_write_options,
    AnnData, AnnDataSet, PyDNAMotif, PyDNAMotifScanner, PyDNAMotifTest, concat,
    read, read_mtx, read_dataset, read_motifs,
)

__version__ = version("snapatac2")

__all__ = [
    "pp", "tl", "pl", "ex", "metrics",
    "set_write_options", "get_write_options",
    "AnnData", "AnnDataSet", "concat", "read", "read_mtx", "read_dataset", "read_10x_mtx", 
    "PyDNAMotif", "PyDNAMotifScanner", "PyDNAMotifTest", "read_motifs",
]

import logging
logging.basicConfig(
    stream=stderr,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO, 
)


from __future__ import annotations

from scipy.sparse import vstack
import numpy as np
import scipy.sparse as ss

import snapatac2._snapatac2 as internal

def smooth(
    adata: internal.AnnData | internal.AnnDataSet | ss.spmatrix,
    distances: str | ss.spmatrix | None = None,
    inplace: bool = True,
) -> np.ndarray | None:
    """
    Smoothing.

    Parameters
    ----------
    adata
        AnnData or AnnDataSet object.
    inplace
        Whether to store the result in the anndata object.

    Returns
    -------
    if `inplace=True` it stores Spectral embedding of data in the field
    `adata.obsm["X_spectral"]`,
    `adata.uns["spectral_eigenvalue"]`,
    otherwise it returns the result as a numpy array.
    """
 
    if distances is None: distances = "distances"
    if isinstance(str, distances): distances = adata.obsp[distances] 
    data = adata.X[:] if isinstance(internal.AnnData, adata) or isinstance(internal.AnnDataSet, adata) else adata
    data = data * make_diffuse_operator(distances)
    if inplace:
        adata.X = data
    else:
        return data

def make_diffuse_operator(knn_d, t = 3):
    return make_markov_matrix(knn_d)**t

def make_markov_matrix(knn_d):
    """
    Turn a (knn) distance matrix into a markov matrix.
    """
    rows = []
    for i in range(knn_d.shape[0]):
        row = knn_d.getrow(i)
        # local nearest neighbor for estimating local density
        ka = int(row.nnz / 3)
        # set sigma for each cell i to the distance to its kath nearest neighbor
        # FIXME: corner case where sigma == 0
        sigma = np.sort(row.data)[ka - 1]
        # apply guassian kernel to get the affinity matrix
        row.data = np.exp(-np.square(row.data / sigma))
        rows.append(row)
    affinity = vstack(rows, format="csr")

    # symmetrize the affinity matrix
    affinity += affinity.T
    # make stochastic matrix
    s = np.ravel(affinity.sum(axis=1))
    for i in range(affinity.shape[0]):
        affinity.data[affinity.indptr[i] : affinity.indptr[i+1]] /= s[i]
    return affinity


import numpy as np
from scipy.sparse import issparse
from sklearn.decomposition import TruncatedSVD
from sklearn.preprocessing import StandardScaler
from sklearn.utils.extmath import safe_sparse_dot

def tf_idf(data, scale_factor=100000, idf=None):
    sparse_input = issparse(data)

    if idf is None:
        # add small value in case down sample creates empty feature
        _col_sum = data.sum(axis=0)
        if sparse_input:
            col_sum = _col_sum.A1.astype(np.float32) + 0.00001
        else:
            col_sum = _col_sum.ravel().astype(np.float32) + 0.00001
        idf = np.log(1 + data.shape[0] / col_sum).astype(np.float32)
    else:
        idf = idf.astype(np.float32)

    _row_sum = data.sum(axis=1)
    if sparse_input:
        row_sum = _row_sum.A1.astype(np.float32) + 0.00001
    else:
        row_sum = _row_sum.ravel().astype(np.float32) + 0.00001

    tf = data.astype(np.float32)

    if sparse_input:
        tf.data = tf.data / np.repeat(row_sum, tf.getnnz(axis=1))
        tf.data = np.log1p(np.multiply(tf.data, scale_factor, dtype="float32"))
        tf = tf.multiply(idf)
    else:
        tf = tf / row_sum[:, np.newaxis]
        tf = np.log1p(np.multiply(tf, scale_factor, dtype="float32"))
        tf = tf * idf
    return tf, idf


def top_features_idx(data, n_features):
    """
    Select top features with the highest importance in CCs.

    Parameters
    ----------
    data
        data.shape = (n_cc, total_features)
    n_features
        number of features to select

    Returns
    -------
    features_idx : np.array
    """
    # data.shape = (n_cc, total_features)
    n_cc = data.shape[0]
    n_features_per_dim = n_features * 10 // n_cc
    n_features_per_dim = min(n_features_per_dim, data.shape[1] - 1)

    sample_range = np.arange(n_cc)[:, None]

    # get idx of n_features_per_dim features with the highest absolute loadings
    data = np.abs(data)
    idx = np.argpartition(-data, n_features_per_dim, axis=1)[:, :n_features_per_dim]
    # idx.shape = (n_cc, n_features_per_dim)

    # make sure the order of first n_features_per_dim is ordered by loadings
    idx = idx[sample_range, np.argsort(-data[sample_range, idx], axis=1)]

    for i in range(n_features // n_cc + 1, n_features_per_dim):
        features_idx = np.unique(idx[:, :i].flatten())
        if len(features_idx) > n_features:
            return features_idx
    else:
        features_idx = np.unique(idx[:, :n_features_per_dim].flatten())
        return features_idx


def cca(
    data1,
    data2,
    scale1=True,
    scale2=True,
    n_components=50,
    max_cc_cell=20000,
    chunk_size=50000,
    random_state=0,
    svd_algorithm="randomized",
    k_filter=None,
    n_features=200,
):
    np.random.seed(random_state)
    tf_data1, tf_data2, scaler1, scaler2 = downsample(
        data1=data1,
        data2=data2,
        todense=True,
        scale1=scale1,
        scale2=scale2,
        max_cc_cell=max_cc_cell,
        random_state=random_state,
    )

    # CCA decomposition
    model = TruncatedSVD(n_components=n_components, algorithm=svd_algorithm, random_state=random_state)
    tf_data2_t = tf_data2.T.copy()
    ## TODO: this can be optimized to reduce memory.
    U = model.fit_transform(tf_data1.dot(tf_data2_t))

    # select dimensions with non-zero singular values
    sel_dim = model.singular_values_ != 0
    print("non zero dims", sel_dim.sum())

    V = model.components_[sel_dim].T
    U = U[:, sel_dim] / model.singular_values_[sel_dim]

    # compute ccv feature loading
    if k_filter:
        high_dim_feature = top_features_idx(
            np.concatenate([U, V], axis=0).T.dot(np.concatenate([tf_data1, tf_data2], axis=0)), n_features=n_features
        )
    else:
        high_dim_feature = None

    # transform CC
    if data2.shape[0] > max_cc_cell:
        V = []
        for chunk_start in np.arange(0, data2.shape[0], chunk_size):
            if issparse(data2):
                tmp = data2[chunk_start : (chunk_start + chunk_size)].toarray()
            else:
                tmp = data2[chunk_start : (chunk_start + chunk_size)]
            if scale2:
                tmp = scaler2.transform(tmp)
            V.append(np.dot(np.dot(U.T, tf_data1), tmp.T).T)
        V = np.concatenate(V, axis=0)
        V = V / model.singular_values_[sel_dim]

    if data1.shape[0] > max_cc_cell:
        U = []
        for chunk_start in np.arange(0, data1.shape[0], chunk_size):
            if issparse(data1):
                tmp = data1[chunk_start : (chunk_start + chunk_size)].toarray()
            else:
                tmp = data1[chunk_start : (chunk_start + chunk_size)]
            if scale1:
                tmp = scaler1.transform(tmp)
            U.append(np.dot(tmp, np.dot(model.components_[sel_dim], tf_data2).T))
        U = np.concatenate(U, axis=0)
        U = U / model.singular_values_[sel_dim]

    return U, V, high_dim_feature


def adata_cca(adata, group_col, separate_scale=True, n_components=50, random_state=42):
    groups = adata.obs[group_col].unique()
    if len(groups) != 2:
        raise ValueError(f"CCA only handle 2 groups, " f"adata.obs[{group_col}] has {len(groups)} different groups.")
    group_a, group_b = groups
    a = adata[adata.obs[group_col] == group_a, :].X
    b = adata[adata.obs[group_col] == group_b, :].X

    pc, loading, _ = cca(
        data1=a,
        data2=b,
        scale1=separate_scale,
        scale2=separate_scale,
        n_components=n_components,
        random_state=random_state,
    )
    total_cc = np.concatenate([pc, loading], axis=0)
    adata.obsm["X_cca"] = total_cc
    return


# def incremental_cca(a, b, max_chunk_size=10000, random_state=0):
#     """
#     Perform Incremental CCA by chunk dot product and IncrementalPCA
#
#     Parameters
#     ----------
#     a
#         dask.Array of dataset a
#     b
#         dask.Array of dataset b
#     max_chunk_size
#         Chunk size for Incremental fit and transform, the larger the better as long as MEM is enough
#     random_state
#
#     Returns
#     -------
#     Top CCA components
#     """
#     raise NotImplementedError
#     # TODO PC is wrong
#     pca = dIPCA(n_components=50,
#                 whiten=False,
#                 copy=True,
#                 batch_size=None,
#                 svd_solver='auto',
#                 iterated_power=0,
#                 random_state=random_state)
#
#     # partial fit
#     n_sample = a.shape[0]
#     n_chunks = n_sample // max_chunk_size + 1
#     chunk_size = int(n_sample / n_chunks) + 1
#     for chunk_start in range(0, n_sample, chunk_size):
#         print(chunk_start)
#         X_chunk = a[chunk_start:chunk_start + chunk_size, :].dot(b.T)
#         pca.partial_fit(X_chunk)
#
#     # transform
#     pcs = []
#     for chunk_start in range(0, n_sample, chunk_size):
#         print(chunk_start)
#         X_chunk = a[chunk_start:chunk_start + chunk_size, :].dot(b.T)
#         pc_chunk = pca.transform(X_chunk).compute()
#         pcs.append(pc_chunk)
#     pcs = np.concatenate(pcs)
#
#     # concatenate CCA
#     total_cc = np.concatenate([pcs, pca.components_.T])
#     return total_cc


def lsi_cca(
    data1,
    data2,
    scale_factor=100000,
    n_components=50,
    max_cc_cell=20000,
    chunk_size=50000,
    svd_algorithm="randomized",
    min_cov_filter=5,
    random_state=0,
):
    np.random.seed(random_state)

    # down sample data1 and data2 to run tf_idf and CCA
    if max_cc_cell < data1.shape[0]:
        sel1 = np.sort(np.random.choice(np.arange(data1.shape[0]), max_cc_cell, False))
        tf_data1 = data1[sel1, :]
    else:
        tf_data1 = data1
    if max_cc_cell < data2.shape[0]:
        sel2 = np.sort(np.random.choice(np.arange(data2.shape[0]), max_cc_cell, False))
        tf_data2 = data2[sel2, :]
    else:
        tf_data2 = data2

    # filter bin to make sure the min_cov_filter is satisfied
    col_sum1 = tf_data1.sum(axis=0).A1
    col_sum2 = tf_data2.sum(axis=0).A1
    # the same bin_filter will also be used
    # in the chunk transfer below
    bin_filter = np.logical_and(col_sum1 > min_cov_filter, col_sum2 > min_cov_filter)
    tf1, idf1 = tf_idf(tf_data1[:, bin_filter], scale_factor=scale_factor)
    tf2, idf2 = tf_idf(tf_data2[:, bin_filter], scale_factor=scale_factor)

    # CCA part
    model = TruncatedSVD(n_components=n_components, algorithm=svd_algorithm, random_state=0)
    tf = tf1.dot(tf2.T)
    U = model.fit_transform(tf)

    # select non-zero singular values
    # transform the whole dataset 2 to get V
    sel_dim = model.singular_values_ != 0
    nnz_singular_values = model.singular_values_[sel_dim]
    nnz_components = model.components_[sel_dim]
    if max_cc_cell > data2.shape[0]:
        V = nnz_components.T
    else:
        # use the safe_sparse_dot to avoid memory error
        # safe_sparse_dot take both sparse and dense matrix,
        # for dense matrix, it just uses normal numpy dot product
        V = np.concatenate(
            [
                safe_sparse_dot(
                    safe_sparse_dot(U.T[sel_dim], tf1),
                    tf_idf(
                        data2[chunk_start : (chunk_start + chunk_size)][:, bin_filter],
                        scale_factor=scale_factor,
                        idf=idf2,
                    )[
                        0
                    ].T,  # [0] is the tf
                ).T
                for chunk_start in np.arange(0, data2.shape[0], chunk_size)
            ],
            axis=0,
        )
        V = V / np.square(nnz_singular_values)

    # transform the whole dataset 1 to get U
    if max_cc_cell > data1.shape[0]:
        U = U[:, sel_dim] / nnz_singular_values
    else:
        U = np.concatenate(
            [
                safe_sparse_dot(
                    tf_idf(
                        data1[chunk_start : (chunk_start + chunk_size)][:, bin_filter],
                        scale_factor=scale_factor,
                        idf=idf1,
                    )[
                        0
                    ],  # [0] is the tf
                    safe_sparse_dot(nnz_components, tf2).T,
                )
                for chunk_start in np.arange(0, data1.shape[0], chunk_size)
            ],
            axis=0,
        )
        U = U / nnz_singular_values
    return U, V


class LSI:
    def __init__(
        self,
        scale_factor=100000,
        n_components=100,
        algorithm="arpack",
        random_state=0,
        idf=None,
        model=None,
    ):
        self.scale_factor = scale_factor
        if idf is not None:
            self.idf = idf.copy()
        if idf is not None:
            self.model = model
        else:
            self.model = TruncatedSVD(n_components=n_components, algorithm=algorithm, random_state=random_state)

    def fit(self, data):
        tf, idf = tf_idf(data, self.scale_factor)
        self.idf = idf.copy()
        n_rows, n_cols = tf.shape
        self.model.n_components = min(n_rows, n_cols, self.model.n_components)
        self.model.fit(tf)
        return self

    def fit_transform(self, data):
        tf, idf = tf_idf(data, self.scale_factor)
        self.idf = idf.copy()
        n_rows, n_cols = tf.shape
        self.model.n_components = min(n_rows, n_cols, self.model.n_components)
        tf_reduce = self.model.fit_transform(tf)
        return tf_reduce / self.model.singular_values_

    def transform(self, data, chunk_size=50000, scaler=None):
        tf_reduce = []
        for chunk_start in np.arange(0, data.shape[0], chunk_size):
            tf, _ = tf_idf(data[chunk_start : (chunk_start + chunk_size)], self.scale_factor, self.idf)
            tf_reduce.append(self.model.transform(tf))
        return np.concatenate(tf_reduce, axis=0) / self.model.singular_values_


class SVD:
    def __init__(
        self,
        n_components=100,
        algorithm="randomized",
        random_state=0,
    ):
        self.model = TruncatedSVD(n_components=n_components, algorithm=algorithm, random_state=random_state)

    def fit(self, data):
        self.model.fit(data)
        return self

    def fit_transform(self, data):
        return self.model.fit_transform(data)

    def transform(self, data, chunk_size=50000, scaler=None):
        tf_reduce = []
        for chunk_start in np.arange(0, data.shape[0], chunk_size):
            if issparse(data):
                tmp = data[chunk_start : (chunk_start + chunk_size)].toarray()
            else:
                tmp = data[chunk_start : (chunk_start + chunk_size)]
            if scaler:
                tmp = scaler.transform(tmp)
            tf_reduce.append(self.model.transform(tmp))
        return np.concatenate(tf_reduce, axis=0)


def downsample(data1, data2, scale1, scale2, todense, max_cc_cell=20000, random_state=0):
    scaler1, scaler2 = [None, None]
    np.random.seed(random_state)
    if data1.shape[0] > max_cc_cell:
        sel1 = np.random.choice(np.arange(data1.shape[0]), min(max_cc_cell, data1.shape[0]), False)
        tf1 = data1[sel1]
    else:
        tf1 = data1.copy()
    if todense:
        if issparse(tf1):
            tf1 = tf1.toarray()

    if data2.shape[0] > max_cc_cell:
        sel2 = np.random.choice(np.arange(data2.shape[0]), min(max_cc_cell, data2.shape[0]), False)
        tf2 = data2[sel2]
    else:
        tf2 = data2.copy()
    if todense:
        if issparse(tf2):
            tf2 = tf2.toarray()

    if scale1:
        scaler1 = StandardScaler()
        tf1 = scaler1.fit_transform(tf1)
    if scale2:
        scaler2 = StandardScaler()
        tf2 = scaler2.fit_transform(tf2)
    return tf1, tf2, scaler1, scaler2



from __future__ import annotations

import numpy as np
import logging

import snapatac2
from snapatac2._snapatac2 import AnnData, AnnDataSet
from snapatac2.tools._misc import aggregate_X
from snapatac2._utils import find_elbow, is_anndata
from ._base import render_plot, heatmap, kde2d, scatter, scatter3d
from ._network import network_scores, network_edge_stat
import snapatac2._snapatac2 as internal

__all__ = [
    'tsse', 'frag_size_distr', 'umap', 'network_scores', 'spectral_eigenvalues',
    'regions', 'motif_enrichment', 'coverage'
]

def valid_cells(
    values,
    width: int = 500,
    height: int = 400,
    **kwargs,
):
    import plotly.graph_objects as go

    values = sorted(values, reverse=True)
    result = {}
    for x, y in enumerate(values):
        x = x + 1
        if y in result:
            x_, n = result[y]
            result[y] = (x_ + x, n + 1)
        else:
            result[y] = (x, 1)
    for y, (x, n) in result.items():
        result[y] = x / n
    y, x = zip(*result.items())

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=x, y=y))
    fig.update_xaxes(type="log")
    fig.update_yaxes(type="log")
    fig.update_layout(
        xaxis_title="Barcodes",
        yaxis_title="Counts",
    )

    return render_plot(fig, width, height, **kwargs)

def tsse(
    adata: AnnData,
    min_fragment: int = 500,
    width: int = 500,
    height: int = 400,
    **kwargs,
) -> 'plotly.graph_objects.Figure' | None:
    """Plot the TSS enrichment vs. number of fragments density figure.

    Parameters
    ----------
    adata
        Annotated data matrix.
    min_fragment
        The cells' unique fragments lower than it should be removed
    width
        The width of the plot
    height
        The height of the plot
    kwargs        
        Additional arguments passed to :func:`~snapatac2.pl.render_plot` to
        control the final plot output. Please see :func:`~snapatac2.pl.render_plot`
        for details.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.

    See Also
    --------
    render_plot

    Examples
    --------
    .. plotly::

        >>> import snapatac2 as snap
        >>> data = snap.read(snap.datasets.pbmc5k(type='h5ad'))
        >>> fig = snap.pl.tsse(data, show=False, out_file=None)
        >>> fig.show()
    """
    if "tsse" not in adata.obs:
        raise ValueError("TSS enrichment score is not computed, please run `metrics.tsse` first.")

    selected_cells = np.where(adata.obs["n_fragment"] >= min_fragment)[0]
    x = adata.obs["n_fragment"].to_numpy()[selected_cells]
    y = adata.obs["tsse"].to_numpy()[selected_cells]

    fig = kde2d(x, y, log_x=True, log_y=False)
    fig.update_layout(
        xaxis_title="Number of unique fragments",
        yaxis_title="TSS enrichment score",
    )

    return render_plot(fig, width, height, **kwargs)

def frag_size_distr(
    adata: AnnData | np.ndarray,
    use_rep: str = "frag_size_distr",
    max_recorded_size: int = 1000,
    **kwargs,
) -> 'plotly.graph_objects.Figure' | None:
    """ Plot the fragment size distribution.
    """
    import plotly.graph_objects as go

    if is_anndata(adata):
        if use_rep not in adata.uns or len(adata.uns[use_rep]) <= max_recorded_size:
            logging.info("Computing fragment size distribution...")
            snapatac2.metrics.frag_size_distr(adata, add_key=use_rep, max_recorded_size=max_recorded_size)
        data = adata.uns[use_rep]
    else:
        data = adata
    data = data[:max_recorded_size+1]

    x, y = zip(*enumerate(data))
    # Make a line plot
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=x[1:], y=y[1:], mode='lines'))
    fig.update_layout(
        xaxis_title="Fragment size",
        yaxis_title="Count",
    )
    return render_plot(fig, **kwargs)

def spectral_eigenvalues(
    adata: AnnData,
    width: int = 600,
    height: int = 400,
    show: bool = True,
    interactive: bool = True,
    out_file: str | None = None,
) -> 'plotly.graph_objects.Figure' | None:
    """Plot the eigenvalues of spectral embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    width
        The width of the plot
    height
        The height of the plot
    show
        Show the figure.
    interactive
        Whether to make interactive plot
    out_file
        Path of the output file for saving the output image, end with
        '.svg' or '.pdf' or '.png' or '.html'.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.
    """
 
    import plotly.express as px
    import pandas as pd

    data = adata.uns["spectral_eigenvalue"]

    df = pd.DataFrame({"Component": map(str, range(1, data.shape[0] + 1)), "Eigenvalue": data})
    fig = px.scatter(df, x="Component", y="Eigenvalue", template="plotly_white")
    n = find_elbow(data)
    adata.uns["num_eigen"] = n
    fig.add_vline(x=n)

    return render_plot(fig, width, height, interactive, show, out_file)

def regions(
    adata: AnnData | AnnDataSet,
    groupby: str | list[str],
    peaks: dict[str, list[str]],
    width: float = 600,
    height: float = 400,
    show: bool = True,
    interactive: bool = True,
    out_file: str | None = None,
) -> 'plotly.graph_objects.Figure' | None:
    """
    Parameters
    ----------
    adata
        Annotated data matrix.
    groupby
        Group the cells into different groups. If a `str`, groups are obtained from
        `.obs[groupby]`.
    peaks
        Peaks of each group.
    width
        The width of the plot
    height
        The height of the plot
    show
        Show the figure
    interactive
        Whether to make interactive plot
    out_file
        Path of the output file for saving the output image, end with
        '.svg' or '.pdf' or '.png' or '.html'.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.
    """
    import polars as pl
    import plotly.graph_objects as go

    peaks = np.concatenate([[x for x in p] for p in peaks.values()])
    n = len(peaks)
    if n > 50000:
        logging.warning(f"Input contains {n} peaks, only 50000 peaks will be plotted.")
        np.random.seed(0)
        indices = np.random.choice(n, 50000, replace=False)
        peaks = peaks[sorted(indices)]

    count = aggregate_X(adata, groupby=groupby, normalize="RPKM")
    names = count.obs_names
    count = pl.DataFrame(count.X.T)
    count.columns = list(names)
    idx_map = {x: i for i, x in enumerate(adata.var_names)}
    idx = [idx_map[x] for x in peaks]
    mat = np.log2(1 + count.to_numpy()[idx, :])

    trace = go.Heatmap(
        x=count.columns,
        y=peaks,
        z=mat,
        type='heatmap',
        colorscale='Viridis',
        colorbar={ "title": "log2(1 + RPKM)" },
    )
    data = [trace]
    layout = {
        "yaxis": { "visible": False, "autorange": "reversed" },
        "xaxis": { "title": groupby },
    }
    fig = go.Figure(data=data, layout=layout)
    return render_plot(fig, width, height, interactive, show, out_file)

def umap(
    adata: AnnData | np.ndarray,
    color: str | np.ndarray | None = None,
    use_rep: str = "X_umap",
    marker_size: float = None,
    marker_opacity: float = 1,
    sample_size: int | None = None,
    **kwargs,
) -> 'plotly.graph_objects.Figure' | None:
    """Plot the UMAP embedding.

    Parameters
    ----------
    adata
        Annotated data matrix.
    color
        If the input is a string, it will be used the key to retrieve values from
        `obs`.
    use_rep
        Use the indicated representation in `.obsm`.
    marker_size
        Size of the dots.
    marker_opacity
        Opacity of the dots.
    sample_size
        If the number of cells is larger than `sample_size`, a random sample of
        `sample_size` cells will be used for plotting.
    kwargs        
        Additional arguments passed to :func:`~snapatac2.pl.render_plot` to
        control the final plot output. Please see :func:`~snapatac2.pl.render_plot`
        for details.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.
    """
    from natsort import index_natsorted

    embedding = adata.obsm[use_rep] if is_anndata(adata) else adata
    if isinstance(color, str):
        groups = adata.obs[color].to_numpy()
    else:
        groups = color
        color = "color"
    
    if sample_size is not None and embedding.shape[0] > sample_size:
        idx = np.random.choice(embedding.shape[0], sample_size, replace=False)
        embedding = embedding[idx, :]
        if groups is not None: groups = groups[idx]

    if groups is not None:
        idx = index_natsorted(groups)
        embedding = embedding[idx, :]
        groups = [groups[i] for i in idx]

    if marker_size is None:
        num_points = embedding.shape[0]
        marker_size = (1000 / num_points)**(1/3) * 3

    if embedding.shape[1] >= 3:
        return scatter3d(embedding[:, 0], embedding[:, 1], embedding[:, 2], color=groups,
            x_label="UMAP-1", y_label="UMAP-2", z_label="UMAP-3", color_label=color,
            marker_size=marker_size, marker_opacity=marker_opacity, **kwargs)
    else:
        return scatter(embedding[:, 0], embedding[:, 1], color=groups,
            x_label="UMAP-1", y_label="UMAP-2", color_label=color,
            marker_size=marker_size, marker_opacity=marker_opacity, **kwargs)

def motif_enrichment(
    enrichment: list(str, 'pl.DataFrame'),
    min_log_fc: float = 1,
    max_fdr: float = 0.01,
    **kwargs,
) -> 'plotly.graph_objects.Figure' | None:
    """Plot the motif enrichment result.

    Parameters
    ----------
    enrichment
        Motif enrichment result.
    min_log_fc
        Retain motifs that satisfy: log2-fold-change >= `min_log_fc`.
    max_fdr
        Retain motifs that satisfy: FDR <= `max_fdr`.
    kwargs        
        Additional arguments passed to :func:`~snapatac2.pl.render_plot` to
        control the final plot output. Please see :func:`~snapatac2.pl.render_plot`
        for details.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.
    """
 
    import pandas as pd
    
    fc = np.vstack([df['log2(fold change)'] for df in enrichment.values()])
    filter1 = np.apply_along_axis(lambda x: np.any(np.abs(x) >= min_log_fc), 0, fc)
    
    fdr = np.vstack([df['adjusted p-value'] for df in enrichment.values()])
    filter2 = np.apply_along_axis(lambda x: np.any(x <= max_fdr), 0, fdr)

    passed = np.logical_and(filter1, filter2)
    
    sign = np.sign(fc[:, passed])
    pvals = np.vstack([df['p-value'].to_numpy()[passed] for df in enrichment.values()])
    minval = np.min(pvals[np.nonzero(pvals)])
    pvals = np.clip(pvals, minval, None)
    pvals = sign * np.log(-np.log10(pvals))

    df = pd.DataFrame(
        pvals.T,
        columns=list(enrichment.keys()),
        index=next(iter(enrichment.values()))['id'].to_numpy()[passed],
    )
      
    return heatmap(
        df.to_numpy(),
        row_names=df.index,
        column_names=df.columns,
        colorscale='RdBu_r',
        **kwargs,
    )

def coverage(
    adata: AnnData,
    region: str,
    groupby: str | list[str],
    out_file: str | None = None,
):
    """
    Plot the coverage tracks for different groups of cells. This function requires
    `matplotlib` to be installed.

    This is a simple implementation for quick visualization. For more advanced
    visualization, please consider exporting the data using 
    :func:`~snapatac2.ex.export_coverage` and visualize it using a genome browser.

    Parameters
    ----------
    adata
        Annotated data matrix.
    region
        Genomic region, e.g. 'chr1:100000-200000'.
    groupby
        Group the cells into different groups. If a `str`, groups are obtained from
        `.obs[groupby]`.
    out_file
        Path of the output file for saving the output image.

    See Also
    --------
    export_coverage
    :func:`~snapatac2.ex.export_coverage` 
    """

    from matplotlib import pyplot as plt

    groupby = adata.obs[groupby] if isinstance(groupby, str) else groupby
    groupby = [x for x in groupby]
    signal_values = []
    track_names = []
    for k, v in sorted(list(internal.get_coverage(adata, region, groupby).items())):
        track_names.append(k)
        signal_values.append(v)
    signal_values = np.array(signal_values)

    start, end = region.split(":")[1].split("-")
    start = int(start)
    end = int(end)
    height_per_track = 1.2
    width = 6

    n_tracks, n_points = signal_values.shape
    fig, axes = plt.subplots(
        n_tracks, 1,
        figsize=[width, n_tracks * height_per_track],
        sharex=True,
        constrained_layout=True,
    )

    if n_tracks == 1:
        axes = [axes]

    # Compute global max for y-axis scaling
    global_max = signal_values.max()

    cmap = plt.get_cmap("tab10")
    for i, (ax, signal) in enumerate(zip(axes, signal_values)):
        color = cmap(i % 10)   # cycle through colors if >10 tracks
        ax.fill_between(range(n_points), 0, signal, color=color)
        ax.set_title(track_names[i], fontsize=7)
        ax.spines[['top', 'right']].set_visible(False)
        ax.set_ylim(0, global_max)

    axes[-1].set_xticks([0, n_points - 1])
    axes[-1].set_xticklabels([str(start), str(end)])
    axes[-1].set_xlabel(region)

    fig.supylabel("RPM")

    if out_file is None:
        plt.show()
    else:
        plt.savefig(out_file, dpi=300, bbox_inches='tight')


from __future__ import annotations

import numpy as np
import itertools

import snapatac2._snapatac2 as internal
from snapatac2._utils import is_anndata 

def scanorama_integrate(
    adata: internal.AnnData | internal.AnnDataSet | np.adarray,
    *,
    batch: str | list[str],
    n_neighbors: int = 20,
    use_rep: str = 'X_spectral',
    use_dims: int | list[int] | None = None,
    groupby: str | list[str] | None = None,
    key_added: str | None = None,
    sigma: float = 15,
    approx: bool = True,
    alpha: float = 0.10,
    batch_size: int = 5000,
    inplace: bool = True,
    **kwargs,
):
    """
    Use Scanorama [Hie19]_ to integrate different experiments.

    Scanorama [Hie19]_ is an algorithm for integrating single-cell
    data from multiple experiments stored in an AnnData object. This
    function should be run after performing `tl.spectral` but before computing
    the neighbor graph, as illustrated in the example below.

    This uses the implementation of `scanorama
    <https://github.com/brianhie/scanorama>`__ [Hie19]_.

    Parameters
    ----------
    data
        Matrice or AnnData object. Matrices should be shaped like n_obs x n_vars.
    batch
        Batch labels for cells. If a string, labels will be obtained from `obs`.
    n_neighbors
        Number of mutual nearest neighbors.
    use_rep
        Use the indicated representation in `.obsm`.
    use_dims
        Use these dimensions in `use_rep`.
    groupby
        If specified, split the data into groups and perform batch correction
        on each group separately.
    key_added
        If specified, add the result to ``adata.obsm`` with this key. Otherwise,
        it will be stored in ``adata.obsm[use_rep + "_scanorama"]``.
    inplace
        Whether to store the result in the anndata object.

    Returns
    -------
    np.ndarray | None
        if `inplace=True` it updates adata with the field
        ``adata.obsm[`use_rep`_scanorama]``, containing adjusted principal components.
        Otherwise, it returns the result as a numpy array.
    
    See Also
    --------
    :func:`~snapatac2.tl.spectral`: compute spectral embedding of the data matrix.

    Example
    -------
    First, load libraries and example dataset, and preprocess.

    >>> import snapatac2 as snap
    >>> adata = snap.read(snap.datasets.pbmc5k(type='h5ad'), backed=None)
    >>> snap.pp.select_features(adata)
    >>> snap.tl.spectral(adata)

    We now arbitrarily assign a batch metadata variable to each cell
    for the sake of example, but during real usage there would already
    be a column in ``adata.obs`` giving the experiment each cell came
    from.

    >>> adata.obs['batch'] = 2218*['a'] + 2218*['b']

    Finally, run Scanorama. Afterwards, there will be a new table in
    ``adata.obsm`` containing the Scanorama embeddings.

    >>> snap.pp.scanorama_integrate(adata, batch='batch')
    >>> 'X_spectral_scanorama' in adata.obsm
    True
    """
    if is_anndata(adata):
        mat = adata.obsm[use_rep]
    else:
        mat = adata
        inplace = False

    # Use only the specified dimensions
    if isinstance(use_dims, int): use_dims = range(use_dims) 
    mat = mat if use_dims is None else mat[:, use_dims]

    if isinstance(batch, str):
        batch = adata.obs[batch]

    if groupby is None:
        mat = _scanorama(mat, batch, n_neighbors, sigma, approx, alpha, batch_size, **kwargs)
    else:
        if isinstance(groupby, str): groupby = adata.obs[groupby]
        groups = list(set(groupby))
        for group in groups:
            group_idx = [i for i, x in enumerate(groupby) if x == group]
            mat[group_idx, :] = _scanorama(
                mat[group_idx, :], batch[group_idx], n_neighbors, sigma, approx, alpha, batch_size, **kwargs)

    if inplace:
        if key_added is None:
            adata.obsm[use_rep + "_scanorama"] = mat
        else:
            adata.obsm[key_added] = mat
    else:
        return mat

def _scanorama(data_matrix, batch_labels, knn, sigma, approx, alpha, batch_size, **kwargs):
    try:
        import scanorama
    except ImportError:
        raise ImportError("\nplease install Scanorama:\n\n\tpip install scanorama")
    import pandas as pd

    label_uniq = list(set(batch_labels))

    if len(label_uniq) > 1:
        batch_idx = []
        data_by_batch = []
        for label in label_uniq:
            idx = [i for i, x in enumerate(batch_labels) if x == label]
            batch_idx.append(idx)
            data_by_batch.append(data_matrix[idx,:])
        new_matrix = np.concatenate(scanorama.assemble(
            data_by_batch,
            knn=knn,
            sigma=sigma,
            approx=approx,
            alpha=alpha,
            ds_names=label_uniq,
            batch_size=batch_size,
            verbose=0,
            **kwargs,
        ))
        idx = list(itertools.chain.from_iterable(batch_idx))
        idx = np.argsort(idx)
        data_matrix = new_matrix[idx, :]
    return data_matrix


from __future__ import annotations

import numpy as np
import rustworkx as rx

from ._base import render_plot

def network_edge_stat(
    network: rx.PyDiGraph,
    **kwargs,
):
    """
    Parameters
    ----------
    network
        Network.
    kwargs        
        Additional arguments passed to :func:`~snapatac2.pl.render_plot` to
        control the final plot output. Please see :func:`~snapatac2.pl.render_plot`
        for details.
    """
    from collections import defaultdict
    import plotly.graph_objects as go

    scores = defaultdict(lambda: defaultdict(lambda: []))

    for fr, to, data in network.edge_index_map().values():
        type = "{} -> {}".format(network[fr].type, network[to].type)
        if data.cor_score is not None:
            scores["correlation"][type].append(data.cor_score)
        if data.regr_score is not None:
            scores["regression"][type].append(data.regr_score)
    
    fig = go.Figure()

    for key, vals in scores["correlation"].items():
        fig.add_trace(go.Violin(
            y=vals,
            name=key,
            box_visible=True,
            meanline_visible=True
        ))

    return render_plot(fig, **kwargs)

def network_scores(
    network: rx.PyDiGraph,
    score_name: str,
    width: float = 800,
    height: float = 400,
    show: bool = True,
    interactive: bool = True,
    out_file: str | None = None,
):
    """
    score_name
        Name of the edge attribute
    width
        The width of the plot
    height
        The height of the plot
    """
    import plotly.express as px
    import pandas as pd
    import bisect

    def human_format(num):
        num = float('{:.3g}'.format(num))
        magnitude = 0
        while abs(num) >= 1000:
            magnitude += 1
            num /= 1000.0
        return '{}{}'.format('{:f}'.format(num).rstrip('0').rstrip('.'), ['', 'K', 'M', 'B', 'T'][magnitude])

    break_points = [100, 500, 2000, 20000, 50000, 100000, 500000]
    intervals = []
    for i in range(len(break_points)):
        if i == 0:
            intervals.append("0 - " + human_format(break_points[i]))
        else:
            intervals.append(human_format(break_points[i - 1]) + " - " + human_format(break_points[i]))
    intervals.append("> 500k")
    values = [[] for _ in range(len(intervals))]
    for e in network.edges():
        i = bisect.bisect(break_points, e.distance)
        sc = getattr(e, score_name)
        if sc is not None:
            values[i].append(sc)

    intervals, values = zip(*filter(lambda x: len(x[1]) > 0, zip(intervals, values)))
    values = [np.nanmean(v) for v in values]

    df = pd.DataFrame({
        "Distance to TSS (bp)": intervals,
        "Average score": values,
    })
    fig = px.bar(
        df, x="Distance to TSS (bp)", y="Average score", title = score_name, 
    )
    return render_plot(fig, width, height, interactive, show, out_file)


from __future__ import annotations

import numpy as np
from scipy.stats import gaussian_kde

def scatter(
    X: list[float],
    Y: list[float],
    color: list[float] | None = None,
    x_label: str = "X",
    y_label: str = "Y",
    color_label: str = "Color",
    marker_size: float = 2,
    marker_opacity: float = 0.5,
    **kwargs,
) -> 'plotly.graph_objects.Figure' | None:
    """Plot a scatter plot.

    Parameters
    ----------
    X
        The x coordinates of the points.
    Y
        The y coordinates of the points.
    color
        The color of the points.
    x_label
        The label of the x axis.
    y_label
        The label of the y axis.
    color_label
        The label of the color bar.
    marker_size
        Size of the dots.
    marker_opacity
        Opacity of the dots.
    kwargs        
        Additional arguments passed to :func:`~snapatac2.pl.render_plot` to
        control the final plot output. Please see :func:`~snapatac2.pl.render_plot`
        for details.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.
    """
    import plotly.express as px
    import pandas as pd

    df = pd.DataFrame({
        x_label: X,
        y_label: Y,
    })
    if color is None:
        color_label = None
    else:
        df[color_label] = color

    fig = px.scatter(
        df, x=x_label, y=y_label, color=color_label,
        color_discrete_sequence=px.colors.qualitative.Dark24,
    )
    fig.update_traces(
        marker_size=marker_size,
        marker={"opacity": marker_opacity},
    )
    fig.update_layout(
        template="simple_white",
        legend={'itemsizing': 'constant'},
    )
    return render_plot(fig, **kwargs)

def scatter3d(
    X: list[float],
    Y: list[float],
    Z: list[float],
    color: list[float] | None = None,
    x_label: str = "X",
    y_label: str = "Y",
    z_label: str = "Z",
    color_label: str = "Color",
    marker_size: float = 2,
    marker_opacity: float = 0.5,
    **kwargs,
) -> 'plotly.graph_objects.Figure' | None:
    """Plot a scatter plot.

    Parameters
    ----------
    X
        The x coordinates of the points.
    Y
        The y coordinates of the points.
    Z
        The z coordinates of the points.
    color
        The color of the points.
    x_label
        The label of the x axis.
    y_label
        The label of the y axis.
    z_label
        The label of the z axis.
    color_label
        The label of the color bar.
    marker_size
        Size of the dots.
    marker_opacity
        Opacity of the dots.
    kwargs        
        Additional arguments passed to :func:`~snapatac2.pl.render_plot` to
        control the final plot output. Please see :func:`~snapatac2.pl.render_plot`
        for details.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be 
        returned, which can then be further customized using the plotly API.
    """
    import plotly.express as px
    import pandas as pd

    df = pd.DataFrame({
        x_label: X,
        y_label: Y,
        z_label: Z,
    })
    if color is not None: df[color_label] = color

    fig = px.scatter_3d(
        df, x=x_label, y=y_label, z=z_label, color=color_label,
        color_discrete_sequence=px.colors.qualitative.Dark24,
    )
    fig.update_traces(
        marker_size=marker_size,
        marker={"opacity": marker_opacity},
    )
    fig.update_layout(
        template="simple_white",
        legend={'itemsizing': 'constant'},
    )
    return render_plot(fig, **kwargs)


def heatmap(
    data_array: np.ndarray,
    row_names: list[str] | None = None,
    column_names: list[str] | None = None,
    cluster_columns: bool = True,
    cluster_rows: bool = True,
    colorscale = "Blues",
    linkage: str = "ward",
    **kwargs,
):
    import plotly.graph_objects as go
    import plotly.figure_factory as ff
    import scipy.cluster.hierarchy as sch
    
    link_func = lambda x: sch.linkage(x, linkage)
    fig = go.Figure()
    
    if cluster_columns:
        dendro_upper = ff.create_dendrogram(data_array.T, linkagefun=link_func, orientation='bottom')
        upper_leaves = list(map(int, dendro_upper['layout']['xaxis']['ticktext']))
        data_array = data_array[:, upper_leaves]
        if column_names is not None:
            dendro_upper['layout']['xaxis']['ticktext'] = np.array(column_names)[upper_leaves]
        for i in range(len(dendro_upper['data'])):
            dendro_upper['data'][i]['yaxis'] = 'y2'
            fig.add_trace(dendro_upper['data'][i])
        fig['layout'] = dendro_upper['layout']

    if cluster_rows:
        dendro_side = ff.create_dendrogram(data_array, linkagefun=link_func, orientation='right')
        side_leaves = list(map(int, dendro_side['layout']['yaxis']['ticktext']))
        data_array = data_array[side_leaves, :]
        if row_names is not None:
            dendro_side['layout']['yaxis']['ticktext'] = np.array(row_names)[side_leaves]
        for i in range(len(dendro_side['data'])):
            dendro_side['data'][i]['xaxis'] = 'x2'
            fig.add_trace(dendro_side['data'][i])
        fig['layout']['yaxis'] = dendro_side['layout']['yaxis']

    # Create Heatmap
    heatmap = [go.Heatmap(
        z=data_array,
        colorscale=colorscale,
        colorbar={'orientation': 'h', 'title': 'log(-log(P))'}
    )]
    if cluster_columns:
        heatmap[0]['x'] = dendro_upper['layout']['xaxis']['tickvals']
    if cluster_rows:
        heatmap[0]['y'] = dendro_side['layout']['yaxis']['tickvals']
    for data in heatmap:
        fig.add_trace(data)
        
    fig.update_layout({'showlegend':False, 'hovermode': 'closest'})
    
    if cluster_rows:
        fig.update_layout(
            xaxis={
                'domain': [.15, 1], 'mirror': False, 'showgrid': False,
                'showline': False, 'zeroline': False, 'ticks':"",
            },
            xaxis2={
                'domain': [0, .15], 'mirror': False, 'showgrid': False,
                'showline': False, 'zeroline': False, 'showticklabels': False,
                'ticks':"",
            },
        )
    if cluster_columns:
        fig.update_layout(
            yaxis={
                'domain': [0, .85], 'mirror': False, 'showgrid': False,
                'showline': False, 'zeroline': False, 'showticklabels': True,
                'side': 'right', 'ticks': ""
            },
            yaxis2={
                'domain':[.825, .975], 'mirror': False, 'showgrid': False,
                'showline': False, 'zeroline': False, 'showticklabels': False,
                'ticks':""
            },
        )
    return render_plot(fig, **kwargs)

def render_plot(
    fig: 'plotly.graph_objects.Figure',
    width: int = 600,
    height: int = 400,
    interactive: bool = True,
    show: bool = True,
    out_file: str | None = None,
    scale: float | None = None,
) -> 'plotly.graph_objects.Figure' | None:
    """ Render a plotly figure.

    Note that this function is not intended to be called directly. Instead, it is
    called by other plotting functions in this package.

    Parameters
    ----------
    fig
        The plotly figure to render.
    width
        The width of the plot.
    height
        The height of the plot.
    interactive
        Whether to make interactive plot.
    show
        Show the figure.
    out_file
        Path of the output file for saving the output image, end with
        '.svg' or '.pdf' or '.png' or '.html'.
    scale
        Scale factor for the image. Only used when `out_file` is not None.
        Use scale > 1 to increase resolution and scale < 1 to decrease.

    Returns
    -------
    'plotly.graph_objects.Figure' | None
        If `show=False` and `out_file=None`, an `plotly.graph_objects.Figure` will be
        returned, which can then be further customized using the plotly API.
    """

    fig.update_layout({
        "width": width,
        "height": height,
    })

    # save figure to file
    if out_file is not None:
        if str(out_file).endswith(".html"):
            fig.write_html(out_file, include_plotlyjs="cdn")
        else:
            fig.write_image(out_file, scale=scale)

    # show figure
    if show:
        if interactive:
            fig.show()
        else:
            from IPython.display import Image
            return Image(fig.to_image(format="png"))

    # return plot object
    if not show and not out_file: return fig

def kde2d(
    x: np.ndarray,
    y: np.ndarray,
    log_x: bool = False,
    log_y: bool = False,
) -> 'plotly.graph_objects.Figure' | None:
    """
    Parameters
    ----------
    x
        X coordinates of the data points.
    y
        Y coordinates of the data points.
    log_x
        Whether to use log10(x) as the x-axis.
    log_y
        Whether to use log10(y) as the y-axis.
    """
    import plotly.graph_objects as go
    import plotly.express as px

    if log_x:
        x = np.log10(x)
    if log_y:
        y = np.log10(y)

    '''
    (z, X, Y) = np.histogram2d(x,y, bins=(bins_x, bins_y), density=True)
    z = z.T
    quintiles = np.percentile(np.ravel(z[z>0]), [1,10,20,30,40,50,60,70,80,90])
    z = np.searchsorted(quintiles, np.ravel(z)).reshape(z.shape)
    '''

    estimator = KDE()
    z, (xx, yy) = estimator(x, y)

    levels = _quantile_to_level(z.ravel(), np.linspace(.05, 1, 10))
    z = np.searchsorted(levels, z) + 0.1

    #quintiles = np.percentile(np.ravel(z[z>0]), [5,10,20,30,40,50,60,70,80,90])
    #z = np.searchsorted(quintiles, np.ravel(z)).reshape(z.shape)

    if log_x:
        xx = 10**xx
    if log_y:
        yy = 10**yy
    contour = go.Contour(
        z=z,
        x=xx,
        y=yy,
        line_smoothing=0.85,
        colorscale="Blues",
        colorbar=dict(
            title="Density",
            #thicknessmode="pixels", thickness=50,
            #lenmode="pixels", len=200,
            yanchor="top", y=1,
            tickvals=list(range(1, len(levels))),
            ticktext=["{:.3f}".format(x) for x in levels],
            #ticks="outside", ticksuffix=" bills",
            #dtick=5
        ),
    )
    fig = go.Figure(data=contour)


    #zmax = z.max()
    #breaks = [0.0] + [x / zmax for x in np.quantile(np.ravel(z[z>0]), np.linspace(0.0, 1.0, num=len(colors)-1))]
    #print(breaks)
    #fig.data[0].colorscale = [(a, b[1]) for a, b in zip(breaks, colors)]

    if log_x:
        fig.update_xaxes(type="log")
    if log_y:
        fig.update_yaxes(type="log")
    fig.update_layout(template="simple_white")

    return fig

def _quantile_to_level(data, quantile):
    """Return data levels corresponding to quantile cuts of mass."""
    isoprop = np.asarray(quantile)
    values = np.ravel(data)
    sorted_values = np.sort(values)[::-1]
    normalized_values = np.cumsum(sorted_values) / values.sum()
    idx = np.searchsorted(normalized_values, 1 - isoprop)
    levels = np.take(sorted_values, idx, mode="clip")
    return levels

# Adapted from seaborn
class KDE:
    """Univariate and bivariate kernel density estimator."""
    def __init__(
        self, *,
        bw_method=None,
        bw_adjust=1,
        gridsize=100,
        cut=3,
        clip=None,
        cumulative=False,
    ):
        """Initialize the estimator with its parameters.
        Parameters
        ----------
        bw_method : string, scalar, or callable, optional
            Method for determining the smoothing bandwidth to use; passed to
            :class:`scipy.stats.gaussian_kde`.
        bw_adjust : number, optional
            Factor that multiplicatively scales the value chosen using
            ``bw_method``. Increasing will make the curve smoother. See Notes.
        gridsize : int, optional
            Number of points on each dimension of the evaluation grid.
        cut : number, optional
            Factor, multiplied by the smoothing bandwidth, that determines how
            far the evaluation grid extends past the extreme datapoints. When
            set to 0, truncate the curve at the data limits.
        clip : pair of numbers or None, or a pair of such pairs
            Do not evaluate the density outside of these limits.
        cumulative : bool, optional
            If True, estimate a cumulative distribution function. Requires scipy.
        """
        if clip is None:
            clip = None, None

        self.bw_method = bw_method
        self.bw_adjust = bw_adjust
        self.gridsize = gridsize
        self.cut = cut
        self.clip = clip
        self.cumulative = cumulative
        self.support = None

    def _define_support_grid(self, x, bw, cut, clip, gridsize):
        """Create the grid of evaluation points depending for vector x."""
        clip_lo = -np.inf if clip[0] is None else clip[0]
        clip_hi = +np.inf if clip[1] is None else clip[1]
        gridmin = max(x.min() - bw * cut, clip_lo)
        gridmax = min(x.max() + bw * cut, clip_hi)
        return np.linspace(gridmin, gridmax, gridsize)

    def _define_support_univariate(self, x, weights):
        """Create a 1D grid of evaluation points."""
        kde = self._fit(x, weights)
        bw = np.sqrt(kde.covariance.squeeze())
        grid = self._define_support_grid(
            x, bw, self.cut, self.clip, self.gridsize
        )
        return grid

    def _define_support_bivariate(self, x1, x2, weights):
        """Create a 2D grid of evaluation points."""
        clip = self.clip
        if clip[0] is None or np.isscalar(clip[0]):
            clip = (clip, clip)

        kde = self._fit([x1, x2], weights)
        bw = np.sqrt(np.diag(kde.covariance).squeeze())

        grid1 = self._define_support_grid(
            x1, bw[0], self.cut, clip[0], self.gridsize
        )
        grid2 = self._define_support_grid(
            x2, bw[1], self.cut, clip[1], self.gridsize
        )

        return grid1, grid2

    def define_support(self, x1, x2=None, weights=None, cache=True):
        """Create the evaluation grid for a given data set."""
        if x2 is None:
            support = self._define_support_univariate(x1, weights)
        else:
            support = self._define_support_bivariate(x1, x2, weights)

        if cache:
            self.support = support

        return support

    def _fit(self, fit_data, weights=None):
        """Fit the scipy kde while adding bw_adjust logic and version check."""
        fit_kws = {"bw_method": self.bw_method}
        if weights is not None:
            fit_kws["weights"] = weights

        kde = gaussian_kde(fit_data, **fit_kws)
        kde.set_bandwidth(kde.factor * self.bw_adjust)

        return kde

    def _eval_univariate(self, x, weights=None):
        """Fit and evaluate a univariate on univariate data."""
        support = self.support
        if support is None:
            support = self.define_support(x, cache=False)

        kde = self._fit(x, weights)

        if self.cumulative:
            s_0 = support[0]
            density = np.array([
                kde.integrate_box_1d(s_0, s_i) for s_i in support
            ])
        else:
            density = kde(support)

        return density, support

    def _eval_bivariate(self, x1, x2, weights=None):
        """Fit and evaluate a univariate on bivariate data."""
        support = self.support
        if support is None:
            support = self.define_support(x1, x2, cache=False)

        kde = self._fit([x1, x2], weights)

        if self.cumulative:

            grid1, grid2 = support
            density = np.zeros((grid1.size, grid2.size))
            p0 = grid1.min(), grid2.min()
            for i, xi in enumerate(grid1):
                for j, xj in enumerate(grid2):
                    density[i, j] = kde.integrate_box(p0, (xi, xj))

        else:

            xx1, xx2 = np.meshgrid(*support)
            density = kde([xx1.ravel(), xx2.ravel()]).reshape(xx1.shape)

        return density, support

    def __call__(self, x1, x2=None, weights=None):
        """Fit and evaluate on univariate or bivariate data."""
        if x2 is None:
            return self._eval_univariate(x1, weights)
        else:
            return self._eval_bivariate(x1, x2, weights)


from __future__ import annotations

from typing import Literal
from typeguard import typechecked
from pathlib import Path
import pooch

from snapatac2._snapatac2 import read_motifs, PyDNAMotif

# This is a global variable used to store all datasets. It is initialized only once
# when the data is requested.
_datasets = None

def register_datasets():
    global _datasets
    if _datasets is None:
        _datasets = pooch.create(
            path=pooch.os_cache("snapatac2"),
            base_url="http://renlab.sdsc.edu/kai/public_datasets/",
            env="SNAP_DATA_DIR",  # The user can overwrite the storage path by setting this environment variable.
            # The registry specifies the files that can be fetched
            registry={
                "atac_pbmc_500_fastqs.tar": "sha256:5897a4790d2841eff69c85b4bef2825166dd0cc2587da91f42eeed09000c5f47",
                "atac_pbmc_500.bam": "sha256:2fac56ca45186943a1daf9da71aed42263ad43a9428f2388fa5f3bcf6d2754ff",
                "atac_pbmc_500.tsv.gz": "sha256:196c5d7ee0169957417e9f4d5502abf1667ef99453328f8d290d4a7f3b205c6c",
                "atac_pbmc_500_downsample.tsv.gz": "sha256:6053cf4578a140bfd8ce34964602769dc5f5ec6b25ba4f2db23cdbd4681b0e2f",

                "atac_pbmc_5k.tsv.gz": "sha256:5fe44c0f8f76ce1534c1ae418cf0707ca5ef712004eee77c3d98d2d4b35ceaec",
                "atac_pbmc_5k.h5ad": "sha256:92ae7f185cdec26517fd8d5acb60b2ce92c71e0ace824de35589c6d7942cab06",
                "atac_pbmc_5k_annotated.h5ad": "sha256:592f1551c27d0cfe4d81e7febad624d6b7d3ebf977b0c3ea64e06b3f3d76f078",

                "colon_transverse.tar": "sha256:18c56bf405ec0ef8e0e2ea31c63bf2299f21bcb82c67f46e8f70f8d71c65ae0e",
                "HEA_cCRE.bed.gz": "sha256:d69ae94649201cd46ffdc634852acfccc317196637c1786aba82068618001408",

                "10x-Multiome-Pbmc10k-ATAC.h5ad": "sha256:24d030fb7f90453a0303b71a1e3e4e7551857d1e70072752d7fff9c918f77217",
                "10x-Multiome-Pbmc10k-RNA.h5ad": "sha256:a25327acff48b20b295c12221a84fd00f8f3f486ff3e7bd090fdef241b996a22",
                "pbmc_10k_atac.tsv.gz": "md5:a959ef83dfb9cae6ff73ab0147d547d1",

                # TF motifs
                "cisBP_human.meme": "sha256:8bf995450258e61cb1c535d5bf9656d580eb68ba68893fa36b77d17ee0730579",
                "Meuleman_2020.meme": "sha256:400dd60ca61dc8388aa0942b42c95920aad7f6bedf5324005cee7e84bcf5b6d0",

                # Genome files
                "gencode_v41_GRCh37.gff3.gz": "sha256:df96d3f0845127127cc87c729747ae39bc1f4c98de6180b112e71dda13592673",
                "gencode_v41_GRCh37.fa.gz": "sha256:ac73947d38df63ccb00724520a5c31d880c1ca423702ca7ccb7e6c2182a362d9",
                #"gencode_v41_GRCh37.fa.gz": "sha256:94330d402e53cf39a1fef6c132e2500121909c2dfdce95cc31d541404c0ed39e",
                "gencode_v41_GRCh38.gff3.gz": "sha256:b82a655bdb736ca0e463a8f5d00242bedf10fa88ce9d651a017f135c7c4e9285",
                "gencode_v41_GRCh38.fa.gz": "sha256:4fac949d7021cbe11117ddab8ec1960004df423d672446cadfbc8cca8007e228",
                "gencode_vM25_GRCm38.gff3.gz": "sha256:e8ed48bef6a44fdf0db7c10a551d4398aa341318d00fbd9efd69530593106846",
                "gencode_vM25_GRCm38.fa.gz": "sha256:617b10dc7ef90354c3b6af986e45d6d9621242b64ed3a94c9abeac3e45f18c17",
                "gencode_vM30_GRCm39.gff3.gz": "sha256:6f433e2676e26569a678ce78b37e94a64ddd50a09479e433ad6f75e37dc82e48",
                "gencode_vM30_GRCm39.fa.gz": "sha256:3b923c06a0d291fe646af6bf7beaed7492bf0f6dd5309d4f5904623cab41b0aa",
            },
            urls={
                "atac_pbmc_500_fastqs.tar": "https://cf.10xgenomics.com/samples/cell-atac/2.0.0/atac_pbmc_500_nextgem/atac_pbmc_500_nextgem_fastqs.tar",
                "atac_pbmc_500.tsv.gz": "https://cf.10xgenomics.com/samples/cell-atac/2.0.0/atac_pbmc_500_nextgem/atac_pbmc_500_nextgem_fragments.tsv.gz",
                "atac_pbmc_500.bam": "https://cf.10xgenomics.com/samples/cell-atac/2.0.0/atac_pbmc_500_nextgem/atac_pbmc_500_nextgem_possorted_bam.bam",
                "atac_pbmc_500_downsample.tsv.gz": "https://osf.io/download/wjv4b",

                "atac_pbmc_5k.tsv.gz": "https://cf.10xgenomics.com/samples/cell-atac/2.0.0/atac_pbmc_5k_nextgem/atac_pbmc_5k_nextgem_fragments.tsv.gz", 
                "atac_pbmc_5k.h5ad": "https://osf.io/download/rj9nc/",
                "atac_pbmc_5k_annotated.h5ad": "https://osf.io/download/e9vc3/",

                "10x-Multiome-Pbmc10k-ATAC.h5ad": "https://data.mendeley.com/api/datasets/dr2z4jbcx3/draft/files/165dfb5c-c557-42a0-bd21-1276d4d7b23e?a=758c37e5-4832-4c91-af89-9a1a83a051b3",
                "10x-Multiome-Pbmc10k-RNA.h5ad": "https://data.mendeley.com/api/datasets/dr2z4jbcx3/draft/files/d079a087-2913-4e29-979e-638e5932bd8c?a=758c37e5-4832-4c91-af89-9a1a83a051b3", 
                "pbmc_10k_atac.tsv.gz": "https://cf.10xgenomics.com/samples/cell-arc/1.0.0/pbmc_granulocyte_sorted_10k/pbmc_granulocyte_sorted_10k_atac_fragments.tsv.gz",

                "colon_transverse.tar": "https://data.mendeley.com/api/datasets/dr2z4jbcx3/draft/files/eaa46151-a73f-4ef5-8b05-9648c8d1efda?a=758c37e5-4832-4c91-af89-9a1a83a051b3", 
                "HEA_cCRE.bed.gz": "https://data.mendeley.com/api/datasets/dr2z4jbcx3/draft/files/91f93222-1a24-49a5-92e3-d9105ec53f91?a=758c37e5-4832-4c91-af89-9a1a83a051b3",

                "cisBP_human.meme": "https://osf.io/download/uk6vn",
                "Meuleman_2020.meme": "https://osf.io/download/6uet5/",

                "gencode_v41_GRCh37.gff3.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_41/GRCh37_mapping/gencode.v41lift37.basic.annotation.gff3.gz",
                "gencode_v41_GRCh37.fa.gz": "https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/hg19.fa.gz",
                #"gencode_v41_GRCh37.fa.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_41/GRCh37_mapping/GRCh37.primary_assembly.genome.fa.gz",
                "gencode_v41_GRCh38.gff3.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_41/gencode.v41.basic.annotation.gff3.gz",
                "gencode_v41_GRCh38.fa.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_41/GRCh38.primary_assembly.genome.fa.gz",
                "gencode_vM25_GRCm38.gff3.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_mouse/release_M25/gencode.vM25.basic.annotation.gff3.gz",
                "gencode_vM25_GRCm38.fa.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_mouse/release_M25/GRCm38.primary_assembly.genome.fa.gz",
                "gencode_vM30_GRCm39.gff3.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_mouse/release_M30/gencode.vM30.basic.annotation.gff3.gz",
                "gencode_vM30_GRCm39.fa.gz": "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_mouse/release_M30/GRCm39.primary_assembly.genome.fa.gz",
            },
        )
    return _datasets

@typechecked
def pbmc500(type: Literal['fastq', 'bam', 'fragment'] = 'fragment', downsample: bool = False) -> Path | list[Path]:
    """scATAC-seq dataset of 500 PBMCs from 10x Genomics.

    This function returns the path to the fragment file of the 10X scATAC-seq dataset
    containing ~500 PBMCs.

    Parameters
    ----------
    type
        One of the following:
            - "fragment": the fragment file.
            - "bam": bam file.

    downsampled
        Whether to return downsampled dataset.

    Returns
    -------
    Path | list[Path]
        Path to the fragment file.
    """
    datasets = register_datasets()
    if type == 'fragment':
        if downsample:
            return Path(datasets.fetch("atac_pbmc_500_downsample.tsv.gz", progressbar=True))
        else:
            return Path(datasets.fetch("atac_pbmc_500.tsv.gz", progressbar=True))
    elif type == 'bam':
        return Path(datasets.fetch("atac_pbmc_500.bam", progressbar=True))
    elif type == 'fastq':
        return [Path(f) for f in datasets.fetch("atac_pbmc_500_fastqs.tar", processor=pooch.Untar(), progressbar=True)]

@typechecked
def pbmc5k(type: Literal['fragment', 'h5ad', 'annotated_h5ad'] = 'fragment') -> Path:
    """scATAC-seq dataset of 5k PBMCs from 10x Genomics.

    Parameters
    ----------
    type
        One of the following:
            - "fragment": the fragment file.
            - "h5ad": preprocessed h5ad file.
            - "annotated_h5ad": annotated h5ad file.

    Returns
    -------
    Path
        path to the file.
    """
    datasets = register_datasets()
    if type == "fragment":
        return Path(datasets.fetch("atac_pbmc_5k.tsv.gz", progressbar=True))
    elif type == "h5ad":
        return Path(datasets.fetch("atac_pbmc_5k.h5ad", progressbar=True))
    elif type == "annotated_h5ad":
        return Path(datasets.fetch("atac_pbmc_5k_annotated.h5ad", progressbar=True))

@typechecked
def pbmc10k_multiome(
    modality: Literal['ATAC', 'RNA'] = 'RNA',
    type: Literal['fragment', 'h5ad'] = 'h5ad',
) -> Path:
    """Single-cell multiome dataset of 10k PBMCs from 10x Genomics.

    Parameters
    ----------
    modality
        One of the following:
            - "ATAC": ATAC-seq data.
            - "RNA": RNA-seq data.
    type
        One of the following:
            - "fragment": the fragment file.
            - "h5ad": preprocessed h5ad file.

    Returns
    -------
    Path
        path to the file.
    """
    datasets = register_datasets()
    if modality == 'RNA':
        return Path(datasets.fetch("10x-Multiome-Pbmc10k-RNA.h5ad"))
    elif modality == 'ATAC':
        if type == 'fragment':
            return Path(datasets.fetch("pbmc_10k_atac.tsv.gz"))
        else:
            return Path(datasets.fetch("10x-Multiome-Pbmc10k-ATAC.h5ad"))

def colon() -> list[tuple[str, Path]]:
    """scATAC-seq datasets of five colon transverse samples from [Zhang21]_.

    Returns
    -------
    list[tuple[str, Path]]
        A list of tuples, each tuple contains the sample name and the path to the fragment file.
    """
    files = register_datasets().fetch("colon_transverse.tar", progressbar=True, processor = pooch.Untar())
    return [(fl.split("/")[-1].split("_rep1_fragments")[0], Path(fl)) for fl in files]

def cre_HEA() -> Path:
    """Curated cis-regulatory elements from [Zhang21]_.

    Returns
    -------
    Path
        Path to the gzipped BED file containing the cis-regulatory elements.
    """
    return Path(register_datasets().fetch("HEA_cCRE.bed.gz"))

def cis_bp(unique: bool = True) -> list[PyDNAMotif]:
    """A list of transcription factor motifs curated by the CIS-BP database.

    This function returns motifs curated from the CIS-BP database [Weirauch14]_.
    The motifs can be used to scan the genome for potential binding sites and
    to perform motif enrichment analysis.

    Parameters
    ----------
    unique
        A transcription factor may have multiple motifs. If `unique=True`, 
        only the motifs with the highest information content will be selected.

    Returns
    -------
    list[PyDNAMotif]
        A list of motifs.

    See Also
    --------
    :func:`~snapatac2.tl.motif_enrichment`: compute motif enrichment.
    """
    motifs = read_motifs(register_datasets().fetch("cisBP_human.meme"))
    for motif in motifs:
        motif.name = motif.id.split('+')[0]
    if unique:
        unique_motifs = {}
        for motif in motifs:
            name = motif.name
            if (
                    name not in unique_motifs or 
                    unique_motifs[name].info_content() < motif.info_content()
               ):
               unique_motifs[name] = motif
        motifs = list(unique_motifs.values())
    return motifs

def Meuleman_2020() -> list[PyDNAMotif]:
    """A list of transcription factor motifs curated from [Meuleman20]_.

    This function returns motifs curated from [Meuleman20]_.
    The motifs in this list have been grouped into families.
    The motifs can be used to scan the genome for potential binding sites and
    to perform motif enrichment analysis.

    Returns
    -------
    list[PyDNAMotif]
        A list of motifs.

    See Also
    --------
    :func:`~snapatac2.tl.motif_enrichment`: compute motif enrichment.
    """
    motifs = read_motifs(register_datasets().fetch("Meuleman_2020.meme"))
    for motif in motifs:
        motif.name = motif.id.split('_')[0]
        motif.family = motif.id.split('+')[-1]
    return motifs